from copy import deepcopy
from operator import itemgetter

import six
from collective.solr.interfaces import IFacetTitleVocabularyFactory
from plone.app.layout.viewlets.common import SearchBoxViewlet
from plone.registry.interfaces import IRegistry
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from six.moves.urllib.parse import urlencode
from zope.component import getUtility, queryUtility
from zope.i18n import translate
from zope.i18nmessageid import Message


def param(view, name):
    """return a request parameter as a list"""
    value = view.request.form.get(name, [])
    if isinstance(value, six.string_types):
        value = [value]
    return value


def facetParameters(view):
    """determine facet fields to be queried for"""
    marker = []
    fields = view.request.get("facet.field", view.request.get("facet_field", marker))
    if isinstance(fields, six.string_types):
        fields = [fields]
    if fields is marker:
        fields = getattr(view, "facet_fields", marker)
    if fields is marker:
        fields = getattr(view.context, "facet_fields", marker)
    if fields is marker:
        registry = getUtility(IRegistry)
        fields = registry["collective.solr.facets"]
    dependencies = {}
    for idx, field in enumerate(fields):
        if ":" in field:
            facet, dep = [f.strip() for f in field.split(":", 1)]
            dependencies[facet] = [d.strip() for d in dep.split(",")]
    return fields, dependencies


def convertFacets(fields, view, filter=None):
    """convert facet info to a form easy to process in templates"""
    info = []
    params = view.request.form.copy()
    if "b_start" in params:
        del params["b_start"]  # Clear the batch when limiting a result set
    facets, dependencies = list(facetParameters(view))
    params["facet.field"] = facets = list(facets)
    fq = params.get("fq", [])
    if isinstance(fq, six.string_types):
        fq = params["fq"] = [fq]
    selected = set([facet.split(":", 1)[0] for facet in fq])
    for field, values in fields.items():
        counts = []
        vfactory = queryUtility(IFacetTitleVocabularyFactory, name=field)
        if vfactory is None:
            # Use the default fallback
            vfactory = getUtility(IFacetTitleVocabularyFactory)
        vocabulary = vfactory(view.context)

        sorted_values = sorted(list(values.items()), key=itemgetter(1), reverse=True)
        for name, count in sorted_values:
            p = deepcopy(params)
            p.setdefault("fq", []).append(('%s:"%s"' % (field, name)).encode("utf-8"))
            if field in p.get("facet.field", []):
                p["facet.field"].remove(field)
            if filter is None or filter(name, count):
                title = name
                if name in vocabulary:
                    title = vocabulary.getTerm(name).title
                if isinstance(title, Message):
                    title = translate(title, context=view.request)
                counts.append(
                    dict(
                        name=name,
                        count=count,
                        title=title,
                        query=urlencode(p, doseq=True),
                    )
                )
        deps = dependencies.get(field, None)
        visible = deps is None or selected.intersection(deps)
        if counts and visible:
            info.append(dict(title=field, counts=counts, name=name))
    if facets:  # sort according to given facets (if available)

        def pos(item):
            try:
                return facets.index(item)
            except ValueError:
                return len(facets)  # position the item at the end

        sortkey = pos
    else:  # otherwise sort by title
        sortkey = itemgetter("title")
    return sorted(info, key=sortkey)


class FacetMixin:
    """mixin with helpers common to the viewlet and view"""

    hidden = ViewPageTemplateFile("hiddenfields.pt")

    def hiddenfields(self):
        """render hidden fields suitable for inclusion in search forms"""
        facets, dependencies = facetParameters(self)
        queries = param(self, "fq")
        return self.hidden(facets=facets, queries=queries)


class SearchBox(SearchBoxViewlet, FacetMixin):

    index = ViewPageTemplateFile("searchbox.pt")


class SearchFacetsView(BrowserView, FacetMixin):
    """view for displaying facetting info as provided by solr searches"""

    def __call__(self, *args, **kw):
        self.args = args
        self.kw = kw
        return super(SearchFacetsView, self).__call__(*args, **kw)

    def facets(self):
        """prepare and return facetting info for the given SolrResponse"""
        results = self.kw.get("results", None)
        fcs = getattr(results, "facet_counts", None)
        if results is not None and fcs is not None:
            filter = lambda name, count: name and count > 0
            return convertFacets(fcs.get("facet_fields", {}), self, filter)
        else:
            return None

    def selected(self):
        """determine selected facets and prepare links to clear them;
        this assumes that facets are selected using filter queries"""
        info = []
        facets = param(self, "facet.field")
        fq = param(self, "fq")
        for idx, query in enumerate(fq):
            field, value = query.split(":", 1)
            params = self.request.form.copy()
            params["fq"] = fq[:idx] + fq[idx + 1 :]
            if field not in facets:
                params["facet.field"] = facets + [field]
            if value.startswith('"') and value.endswith('"'):
                # Look up a vocabulary to provide a title for this facet
                vfactory = queryUtility(IFacetTitleVocabularyFactory, name=field)
                if vfactory is None:
                    # Use the default fallback
                    vfactory = getUtility(IFacetTitleVocabularyFactory)
                vocabulary = vfactory(self.context)
                value = value[1:-1]
                if value in vocabulary:
                    value = vocabulary.getTerm(value).title
                if isinstance(value, Message):
                    value = translate(value, context=self.request)

                info.append(
                    dict(title=field, value=value, query=urlencode(params, doseq=True))
                )
        return info



from collective.solr import SolrMessageFactory as _
from plone.autoform.interfaces import IFormFieldProvider
from plone.supermodel import directives, model
from zope import schema
from zope.interface import provider


@provider(IFormFieldProvider)
class ISolrFields(model.Schema):
    """Additional fields to control Solr integration"""

    directives.fieldset("categorization", fields=["showinsearch", "searchwords"])

    showinsearch = schema.Bool(
        required=False,
        default=True,
        missing_value=True,
        title=_("label_showinsearch", default=u"Show in search"),
        description=_(
            "help_showinsearch",
            default="When checked, this content can be found by anybody via search",
        ),
    )

    searchwords = schema.List(
        required=False,
        default=[],
        missing_value=[],
        title=_("label_searchwords", default=u"Search words"),
        value_type=schema.TextLine(),
        description=_(
            "help_searchwords",
            u"Specify words for which this item will show up "
            u"as the first search result. Multiple words can be "
            u"specified on new lines.",
        ),
    )


# EOF



import logging

from collective.solr.interfaces import ISolrConnectionConfig, ISolrSchema
from plone import api
from plone.registry import Record, field
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from zope.component import getSiteManager, getUtility, queryUtility

logger = logging.getLogger("collective.solr")
PROFILE_ID = "profile-collective.solr:default"


def migrateTo2(context):
    setup_tool = getToolByName(context, "portal_setup")
    setup_tool.runImportStepFromProfile(PROFILE_ID, "browserlayer")
    logger.info("Migrated to version 2: add browserlayer")


def update_registry(context):
    registry = getUtility(IRegistry)
    util = queryUtility(ISolrConnectionConfig)
    if util is not None:
        # Preferably we would get the previous data from this utility, so we
        # can restore the settings.  But the utility is broken.  The best we
        # can do, is remove it.
        sm = getSiteManager()
        sm.unregisterUtility(provided=ISolrConnectionConfig)

    registry.registerInterface(ISolrSchema, prefix="collective.solr")


def migrateTo4(context):
    registry = getUtility(IRegistry)
    if "collective.solr.async" in registry.records:
        old_record = registry.records["collective.solr.async"]
        registry.records["collective.solr.async_indexing"] = old_record
        del registry.records["collective.solr.async"]
        logger.info("Migrated to async_indexing setting")

    pt = api.portal.get_tool("portal_types")
    type_ids = (
        "Collection",
        "Document",
        "Event",
        "File",
        "Folder",
        "Image",
        "Link",
        "News Item",
    )
    new_behavior = "collective.solr.behaviors.ISolrFields"
    for type_id in type_ids:
        if type_id not in pt.objectIds():
            continue
        fti = pt[type_id]
        if new_behavior not in fti.behaviors:
            fti.behaviors += (new_behavior,)
            logger.info("Added new behavior to {}".format(type_id))

    logger.info("Migrated to version 4")


def migrate_to_5(context):
    registry = getUtility(IRegistry)
    if "collective.solr.login" not in registry.records:
        registry_field = field.TextLine(title=u"Login")
        registry_record = Record(registry_field)
        registry_record.value = None
        registry.records["collective.solr.login"] = registry_record
        logger.info("Create registry entry for collective.solr.login")
    if "collective.solr.password" not in registry.records:
        registry_field = field.TextLine(title=u"Password")
        registry_record = Record(registry_field)
        registry_record.value = None
        registry.records["collective.solr.password"] = registry_record
        logger.info("Create registry entry for collective.solr.password")
    if "collective.solr.use_tika" not in registry.records:
        registry_field = field.Bool(title=u"Use Tika")
        registry_record = Record(registry_field)
        registry_record.value = False
        registry.records["collective.solr.use_tika"] = registry_record
        logger.info("Create registry entry for collective.solr.use_tika")
    logger.info("Migrated to version 5")


def migrate_to_6(context):
    registry = getUtility(IRegistry)
    old_login = None
    old_password = None

    if "collective.solr.login" in registry.records:
        old_login = registry["collective.solr.login"]
        del registry.records["collective.solr.login"]
        logger.info("Remove registry entry for collective.solr.login")
    if "collective.solr.password" in registry.records:
        old_password = registry["collective.solr.password"]
        del registry.records["collective.solr.password"]
        logger.info("Remove registry entry for collective.solr.password")

    if "collective.solr.solr_login" not in registry.records:
        registry_field = field.TextLine(title=u"Login")
        registry_record = Record(registry_field)
        if old_login:
            registry_record.value = old_login
        else:
            registry_record.value = None
        registry.records["collective.solr.solr_login"] = registry_record
        logger.info("Create registry entry for collective.solr.solr_login")
    if "collective.solr_password" not in registry.records:
        registry_field = field.TextLine(title=u"Password")
        registry_record = Record(registry_field)
        if old_password:
            registry_record.value = old_password
        else:
            registry_record.value = None
        registry.records["collective.solr.solr_password"] = registry_record
        logger.info("Create registry entry for collective.solr.solr_password")

    logger.info("Migrated to version 6")


def migrate_to_7(context):
    registry = getUtility(IRegistry)
    if "collective.solr.tika_default_field" not in registry.records:
        registry_field = field.TextLine(title=u"Tika Default Field")
        registry_record = Record(registry_field)
        registry_record.value = "content"
        registry.records["collective.solr.tika_default_field"] = registry_record
    logger.info("Migrated to version 7")



from unittest import TestCase

from collective.solr.exceptions import SolrConnectionException
from collective.solr.parser import SolrResponse
from collective.solr.testing import LEGACY_COLLECTIVE_SOLR_INTEGRATION_TESTING
from collective.solr.tests.utils import getData
from collective.solr.utils import (
    findObjects,
    isSimpleSearch,
    isSimpleTerm,
    isWildCard,
    padResults,
    prepare_wildcard,
    prepareData,
    removeSpecialCharactersAndOperators,
    setupTranslationMap,
    splitSimpleSearch,
)


class UtilsTests(TestCase):

    layer = LEGACY_COLLECTIVE_SOLR_INTEGRATION_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.app.manage_addFolder(id="portal", title="Portal")
        self.portal = self.app.portal
        self.portal.manage_addFolder(id="foo", title="Foo")
        self.portal.foo.manage_addFolder(id="bar", title="Bar")
        self.portal.foo.bar.manage_addDocument(id="doc1", title="a document")
        self.portal.foo.bar.manage_addDocument(id="file1", title="a file")
        self.portal.manage_addFolder(id="bar", title="Bar")
        self.portal.bar.manage_addFolder(id="foo", title="Foo")
        self.portal.bar.foo.manage_addDocument(id="doc2", title="a document")
        self.portal.bar.foo.manage_addDocument(id="file2", title="a file")
        self.good = (
            "bar",
            "bar/foo",
            "bar/foo/doc2",
            "bar/foo/file2",
            "foo",
            "foo/bar",
            "foo/bar/doc1",
            "foo/bar/file1",
        )

    def ids(self, results):
        return tuple(sorted([r[0] for r in results]))

    def testZopeFindAndApply(self):
        found = self.app.ZopeFindAndApply(self.portal, search_sub=True)
        self.assertEqual(self.ids(found), self.good)

    def testFindObjects(self):
        found = list(findObjects(self.portal))
        # the starting point itself is returned
        self.assertEqual(found[0], ("", self.portal))
        # but the rest should be the same...
        self.assertEqual(self.ids(found[1:]), self.good)

    def testSimpleTerm(self):
        self.assertTrue(isSimpleTerm("foo"))
        self.assertTrue(isSimpleTerm("foo "))
        self.assertTrue(isSimpleTerm(u"føø"))
        self.assertTrue(isSimpleTerm("føø"))
        self.assertFalse(isSimpleTerm("foo!"))
        self.assertFalse(isSimpleTerm('"foo"'))
        self.assertFalse(isSimpleTerm(u"føø!"))
        # XXX Why would this be false?
        # self.assertFalse(isSimpleTerm(six.text_type('föö', 'latin')))
        self.assertTrue(isSimpleTerm("foo42"))
        self.assertTrue(isSimpleTerm("42foo"))
        self.assertFalse(isSimpleTerm("foo bar"))
        self.assertFalse(isSimpleTerm("42 foo"))

    def testSimpleSearch(self):
        self.assertTrue(isSimpleSearch("foo"))
        self.assertTrue(isSimpleSearch("foo bar"))
        self.assertTrue(isSimpleSearch("foo bar "))
        self.assertTrue(isSimpleSearch("foo   bar"))
        self.assertTrue(isSimpleSearch(u"føø bär"))
        self.assertTrue(isSimpleSearch("føø bär"))
        self.assertTrue(isSimpleSearch("foo*"))
        self.assertTrue(isSimpleSearch("foo* bar*"))
        self.assertTrue(isSimpleSearch("*foo*"))
        self.assertTrue(isSimpleSearch('"foo"'))
        self.assertTrue(isSimpleSearch('"foo bar"'))
        self.assertTrue(isSimpleSearch('"foo AND bar"'))
        self.assertTrue(isSimpleSearch('foo "AND" bar'))
        self.assertTrue(isSimpleSearch('"foo" "bar"'))
        self.assertTrue(isSimpleSearch("fo?bar"))
        self.assertTrue(isSimpleSearch("foo bar?"))
        self.assertTrue(
            isSimpleSearch(
                "areallyverylongword " "andanotherreallylongwordwithsomecake"
            )
        )
        self.assertTrue(
            isSimpleSearch(
                "areallyverylongword " "andanotherreallylongwordwithsomecake *"
            )
        )
        self.assertFalse(isSimpleSearch(""))
        self.assertFalse(isSimpleSearch(u"føø bär!"))
        # XXX Why would this be false?
        # self.assertFalse(isSimpleSearch(six.text_type('föö bär', 'latin')))
        self.assertFalse(isSimpleSearch("foo AND bar"))
        self.assertFalse(isSimpleSearch("foo OR bar"))
        self.assertFalse(isSimpleSearch("foo NOT bar"))
        self.assertFalse(isSimpleSearch('"foo" OR bar'))
        self.assertFalse(isSimpleSearch("(foo OR bar)"))
        self.assertFalse(isSimpleSearch("+foo"))
        self.assertFalse(isSimpleSearch("name:foo"))
        self.assertFalse(isSimpleSearch("foo && bar"))
        self.assertTrue(isSimpleSearch("2000"))
        self.assertTrue(isSimpleSearch("foo 2000"))
        self.assertFalse(isSimpleSearch("foo 1/2000"))
        self.assertTrue(isSimpleSearch("foo 42 bar11"))
        self.assertTrue(isSimpleSearch("2000 foo"))

    def testSplitSimpleSearch(self):
        self.assertEqual(splitSimpleSearch("foo bar"), ["foo", "bar"])
        self.assertEqual(
            splitSimpleSearch('foo "bar foobar" baz'), ["foo", '"bar foobar"', "baz"]
        )
        self.assertRaises(AssertionError, splitSimpleSearch, "foo AND bar")
        self.assertEqual(splitSimpleSearch("foo 42"), ["foo", "42"])
        # standalone wildcard stripped
        self.assertEqual(splitSimpleSearch("foo *"), ["foo"])
        self.assertEqual(splitSimpleSearch("foo **"), ["foo"])
        # connected wildcard not stripped
        self.assertEqual(splitSimpleSearch("foo**"), ["foo**"])

    def testIsWildCard(self):
        self.assertTrue(isWildCard("foo*"))
        self.assertTrue(isWildCard("fo?"))
        self.assertTrue(isWildCard("fo?o"))
        self.assertTrue(isWildCard("fo*oo"))
        self.assertTrue(isWildCard("fo?o*"))
        self.assertTrue(isWildCard("*foo"))
        self.assertTrue(isWildCard("*foo*"))
        self.assertTrue(isWildCard("foo* bar"))
        self.assertTrue(isWildCard("foo bar?"))
        self.assertTrue(isWildCard("*"))
        self.assertTrue(isWildCard("?"))
        self.assertTrue(isWildCard(u"føø*"))
        self.assertTrue(isWildCard(u"føø*".encode("utf-8")))
        self.assertTrue(isWildCard(u"*føø*"))
        self.assertFalse(isWildCard("foo"))
        self.assertFalse(isWildCard("fo#o"))
        self.assertFalse(isWildCard("foo bar"))
        self.assertFalse(isWildCard(u"føø"))
        self.assertFalse(isWildCard(u"føø".encode("utf-8")))
        # other characters might be meaningful in solr, but we don't
        # distinguish them properly (yet)
        self.assertFalse(isWildCard("foo#?"))

    def testPrepareWildcard(self):
        self.assertEqual(prepare_wildcard("Foo"), "foo")
        self.assertEqual(prepare_wildcard("and"), "and")
        self.assertEqual(prepare_wildcard("or"), "or")
        self.assertEqual(prepare_wildcard("not"), "not")
        self.assertEqual(prepare_wildcard("Foo and bar"), "foo and bar")
        self.assertEqual(prepare_wildcard("Foo AND Bar"), "foo AND bar")
        self.assertEqual(prepare_wildcard("FOO AND NOT BAR"), "foo AND NOT bar")
        self.assertEqual(prepare_wildcard("Foo OR Bar"), "foo OR bar")
        self.assertEqual(prepare_wildcard("FOO OR NOT BAR"), "foo OR NOT bar")
        self.assertEqual(
            prepare_wildcard("FOO AND BAR OR FOO AND NOT BAR"),
            "foo AND bar OR foo AND NOT bar",
        )

    def test_removeSpecialCharactersAndOperators(self):
        special = [
            "AND",
            "OR",
            "NOT",
            "+",
            "-",
            "&",
            "|",
            "!",
            "(",
            ")",
            "{",
            "}",
            "[",
            "]",
            "^",
            "~",
            "*",
            "?",
            ":",
            "\\",
            "/",
            "]",
        ]
        for character_or_operant in special:
            self.assertEqual(
                removeSpecialCharactersAndOperators(
                    "%s text %s text %s"
                    % (character_or_operant, character_or_operant, character_or_operant)
                ),
                "  text   text  ",
                "Character not removed: %s" % character_or_operant,
            )

    def test_solr_exception(self):
        e = SolrConnectionException(503, "Error happend", "<xml></xml>")

        def test_raise():
            raise e

        self.assertRaisesRegexp(
            SolrConnectionException, "HTTP code=503, reason=Error happend", test_raise
        )
        self.assertEqual(
            repr(e), "HTTP code=503, Reason=Error happend, body=<xml></xml>"
        )


class TranslationTests(TestCase):
    def testTranslationMap(self):
        tm = setupTranslationMap()
        self.assertEqual("\f\a\b".translate(tm), " " * 3)
        self.assertEqual("foo\nbar".translate(tm), "foo\nbar")
        self.assertEqual("foo\n\tbar\a\f\r".translate(tm), "foo\n\tbar  \r")

    def testRemoveControlCharacters(self):
        data = {"SearchableText": "foo\n\tbar\a\f\r"}
        prepareData(data)
        self.assertEqual(data, {"SearchableText": "foo\n\tbar  \r"})

    def testUnicodeSearchableText(self):
        data = {"SearchableText": u"f\xf8\xf8 bar"}
        prepareData(data)
        self.assertEqual(data, {"SearchableText": u"f\xf8\xf8 bar"})


class BatchingHelperTests(TestCase):
    def results(self):
        xml_response = getData("quirky_response.txt")
        response = SolrResponse(xml_response)
        return response.response  # the result set is named 'response'

    def testResult(self):
        results = self.results()
        self.assertEqual(results.numFound, "1204")
        self.assertEqual(len(results), 137)
        self.assertEqual(results[0].UID, "7c31adb20d5eee314233abfe48515cf3")

    def testResultsPadding(self):
        results = self.results()
        padResults(results)
        self.assertEqual(len(results), 1204)
        self.assertEqual(results[0].UID, "7c31adb20d5eee314233abfe48515cf3")
        self.assertEqual(results[137:], [None] * (1204 - 137))

    def testResultsPaddingWithStart(self):
        results = self.results()
        padResults(results, start=50)
        self.assertEqual(len(results), 1204)
        self.assertEqual(results[:50], [None] * 50)
        self.assertEqual(results[50].UID, "7c31adb20d5eee314233abfe48515cf3")
        self.assertEqual(results[187:], [None] * (1204 - 187))



import argparse
import logging
import sys

import transaction
from collective.solr.browser.maintenance import SolrMaintenanceView
from collective.solr.interfaces import ISolrConnectionManager
from collective.solr.utils import activate
from Testing.makerequest import makerequest
from zope.component import queryUtility
from zope.site.hooks import setHooks, setSite

logger = logging.getLogger()


def _get_site(app, args):
    # Zope.Startup.zopectl.ZopeCmd.run_entrypoint promises to pass the entry
    # point's name as the first argument and any further arguments after that,
    # but that does not work with plone.recipe.zope2instance. Using positional
    # arguments therefore is unreliable - resolve to using a flag.
    parser = argparse.ArgumentParser()
    parser.add_argument("--plonesite", help="Name of the Plone site", default=None)
    namespace, unused = parser.parse_known_args(args)
    name = namespace.plonesite
    if name is not None:
        if name not in app:
            logger.error("Specified site '%s' not found in database." % name)
            sys.exit(1)
    else:
        from Products.CMFPlone.Portal import PloneSite

        for k, v in app.items():
            if isinstance(v, PloneSite):
                name = k
                break
    if not name:
        logger.error("No Plone site found in database root.")
        sys.exit(1)
    site = getattr(app, name)
    setHooks()
    setSite(site)
    return site


def _solr_connection():
    manager = queryUtility(ISolrConnectionManager)
    conn = manager.getConnection()
    logger.info("Opened Solr connection to %s" % conn.host)
    return conn


def solr_activate(app, args):
    _get_site(app, args)
    activate(active=True)
    transaction.commit()


def solr_deactivate(app, args):
    _get_site(app, args)
    activate(active=False)
    transaction.commit()


def solr_clear_index(app, args):
    """Removes all data from a Solr index. Equivalent to removing the
    `data/index` directory while Solr is stopped.
    You can optionally specify the id of the Plone site with
    --plonesite <siteid>.
    """
    _get_site(app, args)  # calls setSite so queryUtility works
    conn = _solr_connection()
    conn.deleteByQuery("[* TO *]")
    conn.commit(optimize=True)
    conn.close()


def solr_reindex(app, args):
    """Reindex Solr. This is equivalent to /@@solr-maintenance/reindex, but
    can handle more documents. Using reindex from the browser will stop
    eventually if there are too many documents, leaving the index incomplete.
    You can optionally specify the id of the Plone site with
    --plonesite <siteid>.
    """
    site = makerequest(_get_site(app, args))
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--ignore_exceptions",
        help="Ignore exceptions during reindexing (yes/no)",
        choices=["yes", "no"],
        default="yes",
    )
    namespace, unused = parser.parse_known_args(args)
    ignore_exceptions = namespace.ignore_exceptions == "yes"
    mv = SolrMaintenanceView(site, site.REQUEST)
    mv.reindex(ignore_exceptions=ignore_exceptions)


def solr_cleanup(app, args):
    """Cleanup Solr index. Remove no longer existing items from Solr index.
    This is equivalent to /@@solr-maintenance/cleanup.
    """
    site = makerequest(_get_site(app, args))
    parser = argparse.ArgumentParser()

    namespace, unused = parser.parse_known_args(args)
    mv = SolrMaintenanceView(site, site.REQUEST)
    mv.cleanup()


def solr_sync(app, args):
    """Sync Solr index.
    Index missing objects and remove no longer existing objects from Solr.
    This is equivalent to /@@solr-maintenance/sync.
    """
    site = makerequest(_get_site(app, args))
    parser = argparse.ArgumentParser()

    namespace, unused = parser.parse_known_args(args)
    mv = SolrMaintenanceView(site, site.REQUEST)
    mv.sync()



from collective.solr.interfaces import IZCMLSolrConnectionConfig
from collective.solr.manager import ZCMLSolrConnectionConfig
from zope import schema
from zope.component.zcml import utility
from zope.interface import Interface


class ISolrConnectionConfigDirective(Interface):
    """Directive which registers a Solr connection config"""

    host = schema.ASCIILine(
        title=u"Host",
        description=u"The host name of the Solr instance to be used.",
        required=True,
    )

    port = schema.Int(
        title=u"Port",
        description=u"The port of the Solr instance to be used.",
        required=True,
    )

    base = schema.ASCIILine(
        title=u"Base",
        description=u"The base prefix of the Solr instance to be used.",
        required=True,
    )


def solrConnectionConfigDirective(_context, host, port, base):

    utility(
        _context,
        provides=IZCMLSolrConnectionConfig,
        component=ZCMLSolrConnectionConfig(host, port, base),
    )



import doctest
from unittest import TestSuite

from plone.testing import layered

from collective.solr.testing import LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING

try:
    from Products.CMFPlone import relationhelper  # noqa

    HAS_PLONE6 = True
except ImportError:
    HAS_PLONE6 = False

optionflags = (
    doctest.ELLIPSIS
    | doctest.NORMALIZE_WHITESPACE
    | doctest.REPORT_ONLY_FIRST_FAILURE
    | doctest.IGNORE_EXCEPTION_DETAIL
)


def test_suite():
    suite = TestSuite()
    if HAS_PLONE6:
        # XXX: errors.txt doctest tests won't work with Plone 6 because ZServer is gone
        testfiles = ["configlet.txt"]
    else:
        testfiles = ["errors.txt", "configlet.txt"]

    for testfile in testfiles:
        doc_suite = doctest.DocFileSuite(
            testfile, package="collective.solr.tests", optionflags=optionflags
        )
        layer = layered(doc_suite, layer=LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING)
        suite.addTest(layer)
    return suite



from datetime import datetime

from DateTime import DateTime

try:
    from ZTUtils.Lazy import Lazy, _marker
except ImportError:
    from Products.ZCatalog.Lazy import Lazy
    from Products.ZCatalog.Lazy import _marker

from xml.etree.cElementTree import iterparse

import six
from collective.solr.interfaces import ISolrFlare
from zope.interface import implementer


class AttrDict(dict):
    """a dictionary with attribute access"""

    def __getattr__(self, name):
        """look up attributes in dict"""
        marker = []
        value = self.get(name, marker)
        if value is not marker:
            return value
        else:
            raise AttributeError(name)


@implementer(ISolrFlare)
class SolrFlare(AttrDict):
    """a sol(a)r brain, i.e. a data container for search results"""

    __allow_access_to_unprotected_subobjects__ = True


class SolrResults(list):
    """a list of results returned from solr, i.e. sol(a)r flares"""


def parseDate(value):
    """use `DateTime` to parse a date, but take care of solr 1.4
    stripping away leading zeros for the year representation"""
    if value.find("-") < 4:
        year, rest = value.split("-", 1)
        value = "%04d-%s" % (int(year), rest)
    return DateTime(value)


def parse_date_as_datetime(value):
    if value.find("-") < 4:
        year, rest = value.split("-", 1)
        value = "%04d-%s" % (int(year), rest)
    format = "%Y-%m-%dT%H:%M:%S"
    if "." in value:
        format += ".%fZ"
    else:
        format += "Z"
    return datetime.strptime(value, format)


# unmarshallers for basic types
unmarshallers = {
    "null": lambda x: None,
    "int": int,
    "float": float,
    "double": float,
    "long": int,
    "bool": lambda x: x == "true",
    "str": lambda x: x or "",
    "date": parseDate,
}

# nesting tags along with their factories
nested = {"arr": list, "lst": dict, "result": SolrResults, "doc": SolrFlare}


def setter(item, name, value):
    """sets the named value on item respecting its type"""
    if isinstance(item, list):
        item.append(value)  # name is ignored for lists
    elif isinstance(item, dict):
        item[name] = value
    else:  # object is assumed...
        setattr(item, name, value)


class SolrResponse(Lazy):
    """a solr search response; TODO: this should get an interface!!"""

    __allow_access_to_unprotected_subobjects__ = True

    def __init__(self, data=None, unmarshallers=unmarshallers):
        self.unmarshallers = unmarshallers
        if data is not None:
            self.parse(data)

    def parse(self, data):
        """parse a solr response contained in a string or file-like object"""
        if isinstance(data, six.text_type):
            data = data.encode("utf-8")
        if isinstance(data, six.binary_type):
            data = six.BytesIO(data)
        stack = [self]  # the response object is the outmost container
        elements = iterparse(data, events=("start", "end"))
        for action, elem in elements:
            tag = elem.tag
            if action == "start":
                if tag in nested:
                    data = nested[tag]()
                    for key, value in elem.attrib.items():
                        if not key == "name":  # set extra attributes
                            if isinstance(key, six.binary_type):
                                key = key.decode("utf-8")
                            if isinstance(value, six.binary_type):
                                value = value.decode("utf-8")
                            setattr(data, key, value)
                    stack.append(data)
            elif action == "end":
                if tag in nested:
                    data = stack.pop()
                    setter(stack[-1], elem.get("name"), data)
                elif tag in self.unmarshallers:
                    data = self.unmarshallers[tag](elem.text)
                    setter(stack[-1], elem.get("name"), data)
        return self

    def results(self):
        """return only the list of results, i.e. a `SolrResults` instance"""
        return getattr(self, "response", [])

    @property
    def actual_result_count(self):
        """
        return the actual_result_count
        """
        if self._rlen is not _marker:
            return self._rlen
        if getattr(self, "response", None):
            return int(self.response.numFound)
        return 0

    @actual_result_count.setter
    def actual_result_count(self, value):
        self._rlen = value

    def __len__(self):
        if self._len is not _marker:
            return self._len
        self._len = len(self.results())
        return self._len

    def __getitem__(self, index):
        return self.results()[index]


class SolrField(AttrDict):
    """a schema field representation"""

    def __init__(self, *args, **kw):
        self["required"] = False
        self["multiValued"] = False
        super(SolrField, self).__init__(*args, **kw)


class AttrStr(str):
    """a string class with attributes"""

    def __new__(self, value, **kw):
        return str.__new__(self, value)

    def __init__(self, value, **kw):
        self.__dict__.update(kw)


class SolrSchema(AttrDict):
    """a solr schema parser:  the xml schema is partially parsed and the
    information collected is later on used both for indexing items as
    well as buiding search queries;  for the time being we are mostly
    interested in explicitly defined fields and their data types, so
    all <analyzer> (tokenizers, filters) and <dynamicField> information
    is ignored;  some of the other fields relevant to the implementation,
    like <uniqueKey>, <solrQueryParser> or <defaultSearchField>, are also
    parsed and provided, all others are ignored"""

    def __init__(self, data=None):
        if data is not None:
            self.parse(data)

    def parse(self, data):
        """parse a solr schema to collect information for building
        search and indexing queries later on"""
        if isinstance(data, six.string_types):
            data = six.StringIO(data)
        self["requiredFields"] = required = []
        types = {}
        for action, elem in iterparse(data):
            name = elem.get("name")
            if elem.tag == "fieldType":
                types[name] = elem.attrib
            elif elem.tag == "field":
                field = SolrField(types[elem.get("type")])
                field.update(elem.attrib)
                field["class_"] = field["class"]  # `.class` will not work
                for key, value in field.items():  # convert to `bool`s
                    if value in ("true", "false"):
                        field[key] = value == "true"
                self[name] = field
                if field.get("required", False):
                    required.append(name)
            elif elem.tag in ("uniqueKey", "defaultSearchField"):
                self[elem.tag] = elem.text
            elif elem.tag == "solrQueryParser":
                self[elem.tag] = AttrStr(elem.text, **elem.attrib)

    @property
    def fields(self):
        """return list of all fields the schema consists of"""
        for name, field in self.items():
            if isinstance(field, SolrField):
                yield field

    @property
    def stored(self):
        """return names of all stored fields, a.k.a. metadata"""
        for field in self.fields:
            if field.stored:
                yield field.name



from datetime import date, datetime

import six
from AccessControl import getSecurityManager
from DateTime import DateTime
from plone.registry.interfaces import IRegistry
from pytz import timezone
from six.moves import map
from zope.component import getUtility

from collective.solr.queryparser import quote
from collective.solr.utils import (
    getConfig,
    isSimpleSearch,
    isWildCard,
    prepare_wildcard,
    removeSpecialCharactersAndOperators,
    splitSimpleSearch,
)
from collective.solr.stopword import isStopWord


ranges = {
    "min": "[%s TO *]",
    "max": "[* TO %s]",
    "min:max": "[%s TO %s]",
    "minmax": "[%s TO %s]",
}

sort_aliases = {"sortable_title": "Title"}

query_args = ("range", "operator", "depth")

ignored = ("use_solr", "-C", "solr_complex_search")


def iso8601date(value):
    """convert `date`, `datetime` and `DateTime` to iso 8601 date format"""
    if isinstance(value, date) and not isinstance(value, datetime):
        value = datetime(value.year, value.month, value.day, tzinfo=timezone("UTC"))
    if isinstance(value, datetime):
        value = DateTime(value)
    if isinstance(value, DateTime):
        v = value.toZone("UTC")
        value = "%04d-%02d-%02dT%02d:%02d:%06.3fZ" % (
            v.year(),
            v.month(),
            v.day(),
            v.hour(),
            v.minute(),
            v.second(),
        )
    return value


def makeSimpleExpressions(term, levenstein_distance):
    '''Return a search expression for part of the query that
    includes the levenstein distance and wildcards where appropriate.
    Returns both an expression for "value" and "base_value"'''

    config = getConfig()
    prefix_wildcard = getattr(config, "prefix_wildcard", False)
    prefix_wildcard_str = "*" if prefix_wildcard else ""
    base_value = term
    if levenstein_distance:
        levenstein_expr = "~%s" % levenstein_distance
    else:
        levenstein_expr = ""
    if '"' in term:  # quoted literals
        value = "%s%s" % (term, levenstein_expr)
        base_value = value
    elif isWildCard(term):
        value = prepare_wildcard(term)
        base_value = quote(term.replace("*", "").replace("?", ""))
    else:
        value = "%s%s* OR %s%s" % (
            prefix_wildcard_str,
            prepare_wildcard(term),
            term,
            levenstein_expr,
        )
    return "(%s)" % value, "(%s)" % base_value


def mangleSearchableText(value, config, force_complex_search=False):
    config = config or getConfig()
    pattern = getattr(config, "search_pattern", "")
    force_simple_search = getattr(config, "force_simple_search", False)
    allow_complex_search = getattr(config, "allow_complex_search", False)
    levenstein_distance = getattr(config, "levenshtein_distance", 0)
    prefix_wildcard = getattr(config, "prefix_wildcard", False)
    value_parts = []
    base_value_parts = []

    stripped = value.strip()
    force_complex_search_prefix = False
    if stripped.startswith("solr:"):
        stripped = stripped.replace("solr:", "", 1).strip()
        force_complex_search_prefix = True

    if not force_simple_search and not isSimpleSearch(value):
        return value

    if allow_complex_search and (force_complex_search_prefix or force_complex_search):
        # FIXME: fold in catalog solr_complex_search parameter check
        return stripped

    if force_simple_search:
        value = removeSpecialCharactersAndOperators(value)

    for term in splitSimpleSearch(value):
        (term_value, term_base_value) = makeSimpleExpressions(term, levenstein_distance)
        # If this is a stopword, we never allow an (term AND term*) pattern, just use (term).
        # Otherwise stopwords won't ever be found, because a (stopword*) search will never succeed.
        if isStopWord(term, config):
            term_value = term_base_value
        value_parts.append(term_value)
        base_value_parts.append(term_base_value)

    base_value = " ".join(base_value_parts)
    value = " ".join(value_parts)
    if pattern:
        value = pattern.format(
            value=quote(value, prefix_wildcard=prefix_wildcard), base_value=base_value
        )
        return set([value])  # add literal query parameter
    if pattern:
        pattern = pattern.encode("utf-8")

    return value


def quotePath(path):
    """quote overlap of solr reserved characters and those allowed
    in zope ids (see OFS.ObjectManager.bad_id)"""
    if path.endswith("/"):
        path = path[:-1]
    for reserved in "/-~()":
        path = path.replace(reserved, "\\%s" % reserved)
    return '"%s"' % path


def mangleQuery(keywords, config, schema):
    """translate / mangle query parameters to replace zope specifics
    with equivalent constructs for solr"""
    extras = {}
    force_complex_search = bool(keywords.get("solr_complex_search"))
    for key, value in keywords.copy().items():
        if key.endswith("_usage"):  # convert old-style parameters
            category, spec = value.split(":", 1)
            extras[key[:-6]] = {category: spec}
            del keywords[key]
        elif isinstance(value, dict):  # unify dict parameters
            if "query" in value:
                keywords[key] = value["query"]
                del value["query"]
            else:
                del keywords[key]
            if "not" in value:
                keywords["-%s" % key] = value["not"]
                del value["not"]
            extras[key] = value
        elif getattr(value, "query", None):  # unify object parameters
            keywords[key] = value.query
            extra = dict()
            for arg in query_args:
                arg_val = getattr(value, arg, None)
                if arg_val is not None:
                    extra[arg] = arg_val
            extras[key] = extra
        elif key in ignored:
            del keywords[key]

    # find EPI indexes
    if schema:
        epi_indexes = {}
        for name in schema.keys():
            parts = name.split("_")
            if parts[-1] in ["string", "depth", "parents"]:
                count = epi_indexes.get(parts[0], 0)
                epi_indexes[parts[0]] = count + 1
        epi_indexes = [k for k, v in epi_indexes.items() if v == 3]
    else:
        epi_indexes = ["path"]
    for key, value in keywords.copy().items():
        args = extras.get(key, {})
        if key == "SearchableText":
            keywords[key] = mangleSearchableText(
                value, config, force_complex_search=force_complex_search
            )
            continue
        if key in epi_indexes:
            if isinstance(value, (list, tuple)):
                value = list(map(quotePath, value))
            else:
                value = quotePath(value)
            path = keywords["%s_parents" % key] = value
            del keywords[key]
            if "depth" in args:
                depth = int(args["depth"])
                if depth >= 0:
                    if not isinstance(value, (list, tuple)):
                        path = [path]
                    tmpl = "+(+%s_depth:[%d TO %d] AND +%s_parents:%s)"
                    params = keywords["%s_parents" % key] = set()
                    for p in path:
                        base = len(p.split("/"))
                        params.add(
                            tmpl % (key, base + (depth and 1), base + depth, key, p)
                        )
                del args["depth"]
        elif key == "effectiveRange":
            if isinstance(value, DateTime):
                steps = getattr(config, "effective_steps", 1)
                if steps > 1:
                    value = DateTime(value.timeTime() // steps * steps)
                value = iso8601date(value)
            del keywords[key]
            keywords["effective"] = "[* TO %s]" % value
            keywords["expires"] = "[%s TO *]" % value
        elif key == "show_inactive":
            del keywords[key]  # marker for `effectiveRange`
        elif "range" in args:
            if not isinstance(value, (list, tuple)):
                value = [value]
            payload = list(map(iso8601date, value))
            keywords[key] = ranges[args["range"]] % tuple(payload)
            del args["range"]
        elif "operator" in args:
            if isinstance(value, (list, tuple)) and len(value) > 1:
                sep = " %s " % args["operator"].upper()
                value = sep.join(map(str, list(map(iso8601date, value))))
                keywords[key] = "(%s)" % value
            del args["operator"]
        elif key == "allowedRolesAndUsers":
            if getattr(config, "exclude_user", False):
                token = "user$" + getSecurityManager().getUser().getId()
                if token in value:
                    value.remove(token)
        elif isinstance(value, (DateTime, datetime, date)):
            keywords[key] = iso8601date(value)
        elif not isinstance(value, six.string_types):
            assert not args, "unsupported usage: %r" % args


def subtractQueryParameters(args, request_keywords=None):
    """subtract parameters related to sorting and limiting search results
    from a given set of arguments, also removing them from the input"""

    def get(name):
        for prefix in "sort_", "sort-":
            key = "%s%s" % (prefix, name)
            value = args.get(key, None)
            if value is not None:
                del args[key]
                return value
        return None

    params = {}
    index = get("on")
    if index:
        reverse = get("order") or ""
        reverse = reverse.lower() in ("reverse", "descending")
        order = reverse and "desc" or "asc"
        params["sort"] = "%s %s" % (index, order)

    limit = get("limit")
    if limit:
        params["rows"] = int(limit)

    for key, value in args.copy().items():
        if key in ("fq", "fl", "facet", "hl", "d", "pt", "sfield", "core"):
            params[key] = value
            del args[key]
        elif key.startswith("facet.") or key.startswith("facet_"):
            name = lambda facet: facet.split(":", 1)[0]
            if isinstance(value, list):
                value = list(map(name, value))
            elif isinstance(value, tuple):
                value = tuple(map(name, value))
            else:
                value = name(value)
            params[key.replace("_", ".", 1)] = value
            del args[key]
        elif key == "b_start":
            params["start"] = int(value)
            del args[key]
        elif key == "b_size":
            params["rows"] = int(value)
            del args[key]
        elif key == "request_handler":
            params["request_handler"] = value
            del args[key]

    return params


def cleanupQueryParameters(args, schema):
    """validate and possibly clean up the given query parameters using
    the given solr schema"""
    sort = args.get("sort", None)
    if sort is not None:
        field, order = sort.split(" ", 1)
        if field not in schema:
            field = sort_aliases.get(field, None)
        fld = schema.get(field, None)
        if fld is not None and fld.indexed:
            args["sort"] = "%s %s" % (field, order)
        else:
            del args["sort"]
    if "facet.field" in args and "facet" not in args:
        args["facet"] = "true"
    return args


def optimizeQueryParameters(query, params):
    """optimize query parameters by using filter queries for
    configured indexes"""
    registry = getUtility(IRegistry)
    filter_queries = registry["collective.solr.filter_queries"]
    fq = []
    if filter_queries is not None:
        for idxs in filter_queries:
            idxs = set(idxs.split(" "))
            if idxs.issubset(list(query.keys())):
                fq.append(" ".join([query.pop(idx) for idx in idxs]))
    if "fq" in params:
        if isinstance(params["fq"], list):
            params["fq"].extend(fq)
        else:
            params["fq"] = [params["fq"]] + fq
    elif fq:
        params["fq"] = fq
    if not query:
        query["*"] = "*:*"  # catch all if no regular query is left...



import os
from logging import getLogger
from socket import error

import six
from collective.solr.interfaces import ISolrConnectionManager, IZCMLSolrConnectionConfig
from collective.solr.local import getLocal, setLocal
from collective.solr.solr import SolrConnection
from collective.solr.utils import getConfig, isActive
from plone.registry.interfaces import IRegistry
from six.moves.http_client import CannotSendRequest, ResponseNotReady
from zope.component import getUtility, queryUtility
from zope.interface import implementer

logger = getLogger("collective.solr.manager")
marker = object()


def none_formatter(value):
    """Return `None` if value is considered as empty"""
    if not value:
        return
    return value


@implementer(IZCMLSolrConnectionConfig)
class ZCMLSolrConnectionConfig(object):
    """Connection values that can be configured through zcml"""

    def __init__(self, host, port, base):
        self.host = "%s:%d" % (host, port)
        self.base = base


@implementer(ISolrConnectionManager)
class SolrConnectionManager(object):
    """a thread-local connection manager for solr"""

    lock = False
    core = None

    def __init__(self, active=None):
        if isinstance(active, bool):
            self.setHost(active=active)

    def setCore(self, core):
        self.core = core

    def setHost(self, active=False, host="localhost", port=8983, base="/solr/plone"):
        """set connection parameters"""
        config = getConfig()
        config.active = active
        config.host = six.text_type(host)
        config.port = port
        config.base = six.text_type(base)
        self.closeConnection(clearSchema=True)

    @property
    def connection_key(self):
        if self.core:
            return "connection_{0}".format(self.core)
        return "connection"

    @property
    def schema_key(self):
        if self.core:
            return "schema_{0}".format(self.core)
        return "schema"

    def closeConnection(self, clearSchema=False):
        """close the current connection, if any"""
        logger.debug("closing connection")
        conn = getLocal(self.connection_key)
        if conn is not None:
            conn.close()
            setLocal(self.connection_key, None)
        if clearSchema:
            setLocal(self.schema_key, None)

    def getConfigParameter(self, key, env_key=None, formatter=None):
        """Return config from environment variable or by default from registry"""
        if not env_key:
            # Transform registry to match environment variable key
            # collective.solr.variable become COLLECTIVE_SOLR_VARIABLE
            env_key = key.replace(".", "_").upper()
        registry = getUtility(IRegistry)
        value = os.getenv(env_key, registry[key])
        if formatter:
            return formatter(value)
        return value

    def getConnection(self):
        """returns an existing connection or opens one"""
        if not isActive():
            return None
        conn = getLocal(self.connection_key)
        if conn is not None:
            return conn

        zcmlconfig = queryUtility(IZCMLSolrConnectionConfig)
        config_host = self.getConfigParameter("collective.solr.host")
        config_login = self.getConfigParameter(
            "collective.solr.solr_login",
            env_key="COLLECTIVE_SOLR_LOGIN",
            formatter=none_formatter,
        )
        config_password = self.getConfigParameter(
            "collective.solr.solr_password",
            env_key="COLLECTIVE_SOLR_PASSWORD",
            formatter=none_formatter,
        )
        if zcmlconfig is not None:
            # use connection parameters defined in zcml...
            logger.debug("opening connection to %s", zcmlconfig.host)
            conn = SolrConnection(
                host=zcmlconfig.host,
                solrBase=zcmlconfig.base,
                persistent=True,
                login=config_login,
                password=config_password,
                manager=self,
            )
            setLocal(self.connection_key, conn)
        elif config_host is not None:
            # otherwise use connection parameters defined in control panel...
            config_port = self.getConfigParameter("collective.solr.port", formatter=int)
            config_base = self.getConfigParameter("collective.solr.base")
            host = "%s:%d" % (config_host, config_port)
            logger.debug("opening connection to %s", host)
            conn = SolrConnection(
                host=host,
                solrBase=config_base,
                persistent=True,
                login=config_login,
                password=config_password,
                manager=self,
            )
            setLocal(self.connection_key, conn)
        if self.core:  # Adapt core if it is necessary
            conn.solrBase = "/".join(conn.solrBase.split("/")[:-1] + [self.core])
        return conn

    def getSchema(self):
        """returns the currently used schema or fetches it"""
        schema = getLocal(self.schema_key)
        if schema is None:
            conn = self.getConnection()
            if conn is not None:
                logger.debug("getting schema from solr")
                self.setSearchTimeout()
                try:
                    schema = conn.get_schema()
                    setLocal(self.schema_key, schema)
                except (error, CannotSendRequest, ResponseNotReady):
                    logger.exception("exception while getting schema")
        return schema

    def setTimeout(self, timeout, lock=marker):
        """set the timeout on the current (or to be opened) connection
        to the given value"""
        update = not self.lock  # update if not locked...
        if lock is not marker:
            self.lock = bool(lock)
            update = True  # ...or changed
            logger.debug("%ssetting timeout lock", lock and "" or "re")
        if update:
            conn = self.getConnection()
            if conn is not None:
                logger.debug("setting timeout to %s", timeout)
                conn.setTimeout(timeout)

    def setIndexTimeout(self):
        """set the timeout on the current (or to be opened) connection
        to the value specified for indexing operations"""
        registry = getUtility(IRegistry)
        index_timeout = registry["collective.solr.index_timeout"]
        self.setTimeout(index_timeout or None)

    def setSearchTimeout(self):
        """set the timeout on the current (or to be opened) connection
        to the value specified for search operations"""
        registry = getUtility(IRegistry)
        search_timeout = registry["collective.solr.search_timeout"]
        self.setTimeout(search_timeout or None)



from plone.indexer import indexer
from six.moves import range
from zope.interface import Interface


@indexer(Interface)
def path_string(obj, **kwargs):
    """return physical path as a string"""
    return "/".join(obj.getPhysicalPath())


@indexer(Interface)
def path_depth(obj, **kwargs):
    """return depth of physical path"""
    return len(obj.getPhysicalPath())


@indexer(Interface)
def path_parents(obj, **kwargs):
    """return all parent paths leading up to the object"""
    elements = obj.getPhysicalPath()
    return ["/".join(elements[: n + 1]) for n in range(1, len(elements))]



from unittest import TestCase

from collective.solr.flare import PloneFlare
from collective.solr.parser import SolrFlare


class FlareTests(TestCase):
    def testRelevanceFormatting(self):
        def score(**kw):
            return PloneFlare(SolrFlare(**kw)).data_record_normalized_score_

        self.assertEqual(score(), "n.a.")
        self.assertEqual(score(score=0.04567), "4.6")
        self.assertEqual(score(score="0.04567"), "4.6")
        self.assertEqual(score(score="0.1"), "10.0")



from unittest import TestCase, skipIf

from plone.app.testing import TEST_USER_ID, setRoles
from Products.CMFCore.interfaces import IIndexQueueProcessor
from Products.CMFCore.utils import getToolByName

# from socket import error
# from socket import timeout
# from time import sleep
from transaction import commit
from zope.component import getGlobalSiteManager, getUtilitiesFor, queryUtility
from zope.configuration import xmlconfig
from zope.event import notify
from zope.lifecycleevent import ObjectModifiedEvent

from collective.solr.exceptions import SolrInactiveException
from collective.solr.interfaces import (
    ISearch,
    ISolrConnectionManager,
    ISolrIndexQueueProcessor,
    IZCMLSolrConnectionConfig,
)
from collective.solr.mangler import mangleQuery
from collective.solr.testing import LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING
from collective.solr.tests.utils import fakehttp, getData
from collective.solr.utils import getConfig

try:
    from Products.CMFPlone import relationhelper  # noqa

    HAS_PLONE6 = True
except ImportError:
    HAS_PLONE6 = False


class UtilityTests(TestCase):

    layer = LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING

    def testGenericInterface(self):
        proc = queryUtility(IIndexQueueProcessor, name="solr")
        self.assertTrue(proc, "utility not found")
        self.assertTrue(IIndexQueueProcessor.providedBy(proc))
        self.assertTrue(ISolrIndexQueueProcessor.providedBy(proc))

    def testSolrInterface(self):
        proc = queryUtility(ISolrIndexQueueProcessor, name="solr")
        self.assertTrue(proc, "utility not found")
        self.assertTrue(IIndexQueueProcessor.providedBy(proc))
        self.assertTrue(ISolrIndexQueueProcessor.providedBy(proc))

    def testRegisteredProcessors(self):
        procs = list(getUtilitiesFor(IIndexQueueProcessor))
        self.assertTrue(procs, "no utilities found")
        solr = queryUtility(ISolrIndexQueueProcessor, name="solr")
        self.assertTrue(
            solr in [util for name, util in procs], "solr utility not found"
        )

    def testSearchInterface(self):
        search = queryUtility(ISearch)
        self.assertTrue(search, "search utility not found")
        self.assertTrue(ISearch.providedBy(search))


class QueryManglerTests(TestCase):

    layer = LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING

    def testExcludeUserFromAllowedRolesAndUsers(self):
        config = getConfig()
        search = queryUtility(ISearch)
        schema = search.getManager().getSchema() or {}
        # first test the default setting, i.e. not removing the user
        keywords = dict(allowedRolesAndUsers=["Member", "user$test_user_1_"])
        mangleQuery(keywords, config, schema)
        self.assertEqual(
            keywords, {"allowedRolesAndUsers": ["Member", "user$test_user_1_"]}
        )
        # now let's remove it...
        config.exclude_user = True
        keywords = dict(allowedRolesAndUsers=["Member", "user$test_user_1_"])
        mangleQuery(keywords, config, schema)
        self.assertEqual(keywords, {"allowedRolesAndUsers": ["Member"]})


class IndexingTests(TestCase):

    layer = LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING

    def setUp(self):
        replies = (
            getData("plone_schema.xml"),
            getData("add_response.txt"),
            getData("add_response.txt"),
            getData("add_response.txt"),
            getData("commit_response.txt"),
        )
        self.proc = queryUtility(ISolrConnectionManager)
        self.proc.setHost(active=True)
        conn = self.proc.getConnection()
        fakehttp(conn, *replies)  # fake schema response
        self.proc.getSchema()  # read and cache the schema
        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.portal.invokeFactory("Folder", id="folder")
        self.folder = self.portal.folder
        commit()

    def tearDown(self):
        self.proc.closeConnection(clearSchema=True)
        # due to the `commit()` in the tests below the activation of the
        # solr support in `afterSetUp` needs to be explicitly reversed...
        self.proc.setHost(active=False)
        commit()

    @skipIf(HAS_PLONE6, "ZServer is gone in Plone 6. Therefore this test does not work")
    def testIndexObject(self):
        output = []
        connection = self.proc.getConnection()
        connection.get_schema()  # cache schema to avoid multiple calls

        responses = (getData("plone_schema.xml"), getData("commit_response.txt"))
        output = fakehttp(connection, *responses)  # fake responses
        self.folder.title = "Foo"
        self.portal.invokeFactory("Folder", id="myfolder", title="Foo")
        self.assertEqual(str(output), "", "reindexed unqueued!")
        commit()  # indexing happens on commit
        required = '<field name="Title" update="set">Foo</field>'
        self.assertTrue(str(output).find(required) > 0, '"title" data not found')

    def testNoIndexingForNonCatalogAwareContent(self):
        output = []
        connection = self.proc.getConnection()
        responses = [getData("dummy_response.txt")] * 42  # set up enough...
        output = fakehttp(connection, *responses)  # fake responses
        notify(ObjectModifiedEvent(self.folder))
        self.portal.invokeFactory("Folder", id="myfolder", title="Foo")
        commit()  # indexing happens on commit
        self.assertNotEqual(repr(output).find("Foo"), -1, "title not found")
        self.assertEqual(
            repr(output).find("at_references"), -1, "`at_references` found?"
        )


class SiteSearchTests(TestCase):

    layer = LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]

    def tearDown(self):
        # resetting the solr configuration after each test isn't strictly
        # needed at the moment, but it triggers the `ConnectionStateError`
        # when the other tests (in `errors.txt`) is trying to perform an
        # actual solr search...
        queryUtility(ISolrConnectionManager).setHost(active=False)

    def testInactiveException(self):
        search = queryUtility(ISearch)
        self.assertRaises(SolrInactiveException, search, "foo")


#    def testSearchWithoutServer(self):
#        config = getConfig()
#        config.active = True
#        config.port = 55555     # random port so the real solr might still run
#        search = queryUtility(ISearch)
#        self.assertRaises(error, search, 'foo')

#   Why should this raise a socket error?
#    def testSearchWithoutSearchableTextInPortalCatalog(self):
#        config = queryUtility(ISolrConnectionConfig)
#        config.active = True
#        config.port = 55555     # random port so the real solr might still run
#        catalog = self.portal.portal_catalog
#        catalog.delIndex('SearchableText')
#        self.assertFalse('SearchableText' in catalog.indexes())
#        query = self.portal.restrictedTraverse('queryCatalog')
#        request = dict(SearchableText='foo')
#        self.assertRaises(error, query, request)
#
# XXX: Enabling this test leads to a test deadlock. Needs to be investigated.
# (timo)
#    def testSearchTimeout(self):
#        config = getConfig()
#        config.active = True
#        config.search_timeout = 2.0  # specify the timeout
#        config.port = 55555         # don't let the real solr disturb us
#
#        def quick(handler):         # set up fake http response
#            sleep(0.5)              # and wait a bit before sending it
#            handler.send_response(200, getData('search_response.txt'))
#
#        def slow(handler):          # set up another http response
#            sleep(3)                # but wait longer before sending it
#            handler.send_response(200, getData('search_response.txt'))
#        # We need a third handler, as the second one will timeout, which causes  # noqa
#        # the SolrConnection.doPost method to catch it and try to reconnect.
#        thread = fakeServer([quick, slow, slow], port=55555)
#        search = queryUtility(ISearch)
#        search('foo')               # the first search should succeed
#        try:
#            self.assertRaises(timeout, search, 'foo')   # but not the second
#        finally:
#            thread.join()           # the server thread must always be joined


class ZCMLSetupTests(TestCase):

    layer = LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING

    def tearDown(self):
        manager = queryUtility(ISolrConnectionManager)
        manager.setHost(active=False)
        zcmlconfig = queryUtility(IZCMLSolrConnectionConfig)
        gsm = getGlobalSiteManager()
        gsm.unregisterUtility(zcmlconfig, IZCMLSolrConnectionConfig)

    def testConnectionConfigurationViaZCML(self):
        import collective.solr

        context = xmlconfig.file("meta.zcml", collective.solr)
        xmlconfig.string(
            """
            <configure xmlns:solr="http://namespaces.plone.org/solr">
                <solr:connection host="127.0.0.23" port="3898" base="/foo" />
            </configure>
        """,
            context=context,
        )
        manager = queryUtility(ISolrConnectionManager)
        manager.setHost(active=True)  # also clears connection cache
        connection = manager.getConnection()
        self.assertEqual(connection.host, "127.0.0.23:3898")
        self.assertEqual(connection.solrBase, "/foo")


class SiteSetupTests(TestCase):

    layer = LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]

    def testTranslation(self):
        utrans = getToolByName(self.portal, "translation_service").utranslate

        def translate(msg):
            return utrans(msgid=msg, domain="solr")

        self.assertEqual(translate("portal_type"), "Content type")



import sys

from setuptools import find_packages, setup

version = "9.2.2"

assert sys.version_info >= (
    3,
    6,
    0,
), "collective.solr 9 requires Python 3.8.0+. Please downgrade to collective.solr 8 for Python 2 and Plone 4.3/5.1."


long_description = "\n\n".join([open("README.rst").read(), open("CHANGES.rst").read()])

setup(
    name="collective.solr",
    version=version,
    description="Solr integration for external indexing and searching.",
    long_description=long_description,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 5.2",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Addon",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="plone cmf zope indexing searching solr lucene",
    author="Plone Community",
    author_email="plone-developers@lists.sourceforge.net",
    maintainer="Timo Stollenwerk",
    maintainer_email="tisto@plone.org",
    url="https://github.com/collective/collective.solr",
    license="GPL version 2",
    packages=find_packages("src"),
    package_dir={"": "src"},
    namespace_packages=["collective"],
    include_package_data=True,
    platforms="Any",
    zip_safe=False,
    install_requires=[
        "Acquisition",
        "DateTime",
        "Products.CMFCore",
        "Products.CMFPlone >= 4.3.7",
        "Products.GenericSetup",
        "Products.ZCatalog",
        "lxml",
        "plone.app.content",
        "plone.app.layout",
        "plone.app.registry",
        "plone.app.vocabularies",
        "plone.api",
        "plone.browserlayer",
        "plone.indexer",
        "plone.restapi",
        "requests-toolbelt",
        "setuptools",
        "transaction",
        "zope.component",
        "zope.i18nmessageid",
        "zope.interface",
        "zope.publisher",
        "zope.schema",
    ],
    extras_require={
        "test": ["plone.app.testing[robot]", "plone.app.robotframework[debug]"],
        "test4": [
            "Products.LinguaPlone >=3.1a1",
            "plone.app.robotframework[debug]",
            "plone.app.testing[robot]",
        ],
    },
    entry_points="""
      [z3c.autoinclude.plugin]
      target=plone
      [zopectl.command]
      solr_activate=collective.solr.commands:solr_activate
      solr_deactivate=collective.solr.commands:solr_deactivate
      solr_clear_index=collective.solr.commands:solr_clear_index
      solr_clear=collective.solr.commands:solr_clear_index
      solr_reindex=collective.solr.commands:solr_reindex
      solr_cleanup=collective.solr.commands:solr_cleanup
      solr_sync=collective.solr.commands:solr_sync
    """,
)



from collective.solr.parser import SolrResponse
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.serializer.catalog import LazyCatalogResultSerializer
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface


@implementer(ISerializeToJson)
@adapter(SolrResponse, Interface)
class LazySolrCatalogResultSerializer(LazyCatalogResultSerializer):
    """The LazyCatalog adapter for the plone.restapi JSON serialization"""



from datetime import datetime
from unittest import TestCase

from DateTime import DateTime
from pytz import timezone
from zope.component import getGlobalSiteManager, provideUtility

from collective.solr.interfaces import ISolrConnectionConfig
from collective.solr.mangler import (
    cleanupQueryParameters,
    mangleQuery,
    optimizeQueryParameters,
    subtractQueryParameters,
    mangleSearchableText,
)
from collective.solr.parser import SolrField, SolrSchema
from collective.solr.testing import COLLECTIVE_SOLR_MOCK_REGISTRY_FIXTURE
from collective.solr.utils import getConfig
from unittest import mock


def mangle(**keywords):
    mangleQuery(keywords, None, {})
    return keywords


class Query:
    def __init__(self, query, range=None, operator=None, depth=None):
        self.query = query
        self.range = range
        self.operator = operator
        self.depth = depth


class QueryManglerTests(TestCase):
    layer = COLLECTIVE_SOLR_MOCK_REGISTRY_FIXTURE

    def setUp(self):
        self.config = getConfig()
        provideUtility(self.config, ISolrConnectionConfig)

    def tearDown(self):
        gsm = getGlobalSiteManager()
        gsm.unregisterUtility(self.config, ISolrConnectionConfig)

    def testPassUnknownArguments(self):
        keywords = mangle(foo=23, bar=42)
        self.assertEqual(keywords, {"foo": 23, "bar": 42})
        keywords = mangle(foo=Query(23), bar=Query(42))
        self.assertEqual(keywords, {"foo": 23, "bar": 42})

    def testComplainAboutUnknownUsages(self):
        keywords = dict(foo=23, foo_usage="bar:42")
        self.assertRaises(AssertionError, mangleQuery, keywords, None, {})

    def testMinRange(self):
        keywords = mangle(foo=[23], foo_usage="range:min")
        self.assertEqual(keywords, {"foo": "[23 TO *]"})
        keywords = mangle(foo=dict(query=[23], range="min"))
        self.assertEqual(keywords, {"foo": "[23 TO *]"})
        keywords = mangle(foo=Query(query=[23], range="min"))
        self.assertEqual(keywords, {"foo": "[23 TO *]"})

    def testMaxRange(self):
        keywords = mangle(foo=[23], foo_usage="range:max")
        self.assertEqual(keywords, {"foo": "[* TO 23]"})
        keywords = mangle(foo=dict(query=[23], range="max"))
        self.assertEqual(keywords, {"foo": "[* TO 23]"})
        keywords = mangle(foo=dict(query=23, range="max"))
        self.assertEqual(keywords, {"foo": "[* TO 23]"})
        keywords = mangle(foo=Query(query=23, range="max"))
        self.assertEqual(keywords, {"foo": "[* TO 23]"})

    def testMinMaxRange(self):
        keywords = mangle(foo=(23, 42), foo_usage="range:min:max")
        self.assertEqual(keywords, {"foo": "[23 TO 42]"})
        keywords = mangle(foo=dict(query=(23, 42), range="min:max"))
        self.assertEqual(keywords, {"foo": "[23 TO 42]"})
        keywords = mangle(foo=Query(query=(23, 42), range="min:max"))
        self.assertEqual(keywords, {"foo": "[23 TO 42]"})

    def testDateConversion(self):
        day = DateTime("1972/05/11 UTC")
        keywords = mangle(foo=day)
        self.assertEqual(keywords, {"foo": "1972-05-11T00:00:00.000Z"})
        keywords = mangle(foo=(day, day + 7), foo_usage="range:min:max")
        self.assertEqual(
            keywords, {"foo": "[1972-05-11T00:00:00.000Z TO 1972-05-18T00:00:00.000Z]"}
        )
        keywords = mangle(foo=[day], foo_usage="range:min")
        self.assertEqual(keywords, {"foo": "[1972-05-11T00:00:00.000Z TO *]"})
        keywords = mangle(foo=dict(query=[day], range="min"))
        self.assertEqual(keywords, {"foo": "[1972-05-11T00:00:00.000Z TO *]"})
        keywords = mangle(foo=Query(day))
        self.assertEqual(keywords, {"foo": "1972-05-11T00:00:00.000Z"})

    def testPythonDatetimeConversion(self):
        date = datetime(2020, 1, 13, 9, 23, 5, tzinfo=timezone("CET"))
        keywords = mangle(foo=date)
        self.assertEqual(keywords, {"foo": "2020-01-13T08:23:05.000Z"})

    def testPythonDateConversion(self):
        date = datetime(2020, 1, 15).date()
        keywords = mangle(foo=date)
        self.assertEqual(keywords, {"foo": "2020-01-15T00:00:00.000Z"})

    def testOperatorConversion(self):
        keywords = mangle(foo=(23, 42), foo_usage="operator:or")
        self.assertEqual(keywords, {"foo": "(23 OR 42)"})
        keywords = mangle(foo=dict(query=(23, 42), operator="or"))
        self.assertEqual(keywords, {"foo": "(23 OR 42)"})
        keywords = mangle(foo=Query(query=(23, 42), operator="or"))
        self.assertEqual(keywords, {"foo": "(23 OR 42)"})
        keywords = mangle(foo=(23, 42), foo_usage="operator:and")
        self.assertEqual(keywords, {"foo": "(23 AND 42)"})
        keywords = mangle(foo=dict(query=(23, 42), operator="and"))
        self.assertEqual(keywords, {"foo": "(23 AND 42)"})
        keywords = mangle(foo=Query(query=(23, 42), operator="and"))
        self.assertEqual(keywords, {"foo": "(23 AND 42)"})
        day = DateTime("1972/05/11 UTC")
        keywords = mangle(foo=dict(query=(day, day + 7), operator="or"))
        self.assertEqual(
            keywords, {"foo": "(1972-05-11T00:00:00.000Z OR 1972-05-18T00:00:00.000Z)"}
        )
        keywords = mangle(foo=Query(query=(day, day + 7), operator="or"))
        self.assertEqual(
            keywords, {"foo": "(1972-05-11T00:00:00.000Z OR 1972-05-18T00:00:00.000Z)"}
        )

    def testEffectiveRangeConversion(self):
        day = DateTime("1972/05/11 UTC")
        keywords = mangle(effectiveRange=day, show_inactive=False)
        self.assertEqual(
            keywords,
            {
                "effective": "[* TO 1972-05-11T00:00:00.000Z]",
                "expires": "[1972-05-11T00:00:00.000Z TO *]",
            },
        )

    def testEffectiveRangeSteps(self):
        date = DateTime("1972/05/11 03:47:02 UTC")
        # first test with the default step of 1 seconds, i.e. unaltered
        keywords = mangle(effectiveRange=date, show_inactive=False)
        self.assertEqual(
            keywords,
            {
                "effective": "[* TO 1972-05-11T03:47:02.000Z]",
                "expires": "[1972-05-11T03:47:02.000Z TO *]",
            },
        )
        # and finally with a setting for steps
        self.config.effective_steps = 300
        keywords = dict(effectiveRange=date, show_inactive=False)
        mangleQuery(keywords, self.config, {})
        self.assertEqual(
            keywords,
            {
                "effective": "[* TO 1972-05-11T03:45:00.000Z]",
                "expires": "[1972-05-11T03:45:00.000Z TO *]",
            },
        )

    def testIgnoredParameters(self):
        keywords = mangle(use_solr=True, foo="bar")
        self.assertEqual(keywords, {"foo": "bar"})
        keywords = mangle(**{"-C": True, "foo": "bar"})
        self.assertEqual(keywords, {"foo": "bar"})
        keywords = mangle(**{"solr_complex_search": True, "foo": "bar"})
        self.assertEqual(keywords, {"foo": "bar"})


class PathManglerTests(TestCase):
    def testSimplePathQuery(self):
        keywords = mangle(path="/foo")
        self.assertEqual(keywords, {"path_parents": '"\\/foo"'})

    def testSimplePathQueryAsDictionary(self):
        keywords = mangle(path=dict(query="/foo"))
        self.assertEqual(keywords, {"path_parents": '"\\/foo"'})

    def testSimplePathQueryAsObject(self):
        keywords = mangle(path=Query(query="/foo"))
        self.assertEqual(keywords, {"path_parents": '"\\/foo"'})

    def testPathQueryWithDepth(self):
        keywords = mangle(path=dict(query="/foo", depth=-1))
        self.assertEqual(keywords, {"path_parents": '"\\/foo"'})
        keywords = mangle(path=dict(query="/foo", depth=0))
        self.assertEqual(
            keywords,
            {
                "path_parents": set(
                    ['+(+path_depth:[2 TO 2] AND +path_parents:"\\/foo")']
                )
            },
        )
        keywords = mangle(path=Query(query="/foo", depth=0))
        self.assertEqual(
            keywords,
            {
                "path_parents": set(
                    ['+(+path_depth:[2 TO 2] AND +path_parents:"\\/foo")']
                )
            },
        )
        keywords = mangle(path=dict(query="/foo", depth=2))
        self.assertEqual(
            keywords,
            {
                "path_parents": set(
                    ['+(+path_depth:[3 TO 4] AND +path_parents:"\\/foo")']
                )
            },
        )
        keywords = mangle(path=Query(query="/foo", depth=2))
        self.assertEqual(
            keywords,
            {
                "path_parents": set(
                    ['+(+path_depth:[3 TO 4] AND +path_parents:"\\/foo")']
                )
            },
        )

    def testMultiplePathQuery(self):
        keywords = mangle(path=["/foo", "/bar"])
        self.assertEqual(keywords, {"path_parents": ['"\\/foo"', '"\\/bar"']})
        keywords = mangle(path=dict(query=["/foo", "/bar"]))
        self.assertEqual(keywords, {"path_parents": ['"\\/foo"', '"\\/bar"']})
        keywords = mangle(path=dict(query=["/foo", "/bar"], depth=-1))
        self.assertEqual(keywords, {"path_parents": ['"\\/foo"', '"\\/bar"']})
        keywords = mangle(path=dict(query=["/foo", "/bar"], depth=0))
        self.assertEqual(
            keywords,
            {
                "path_parents": set(
                    [
                        '+(+path_depth:[2 TO 2] AND +path_parents:"\\/foo")',
                        '+(+path_depth:[2 TO 2] AND +path_parents:"\\/bar")',
                    ]
                )
            },
        )
        keywords = mangle(path=dict(query=["/foo", "/bar"], depth=1))
        self.assertEqual(
            keywords,
            {
                "path_parents": set(
                    [
                        '+(+path_depth:[3 TO 3] AND +path_parents:"\\/foo")',
                        '+(+path_depth:[3 TO 3] AND +path_parents:"\\/bar")',
                    ]
                )
            },
        )
        keywords = mangle(path=dict(query=["/a/b", "/c"], depth=-1))
        self.assertEqual(keywords, {"path_parents": ['"\\/a\\/b"', '"\\/c"']})
        keywords = mangle(path=dict(query=["/a/b", "/c"], depth=0))
        self.assertEqual(
            keywords,
            {
                "path_parents": set(
                    [
                        '+(+path_depth:[3 TO 3] AND +path_parents:"\\/a\\/b")',
                        '+(+path_depth:[2 TO 2] AND +path_parents:"\\/c")',
                    ]
                )
            },
        )
        keywords = mangle(path=dict(query=["/a/b", "/c"], depth=2))
        self.assertEqual(
            keywords,
            {
                "path_parents": set(
                    [
                        '+(+path_depth:[4 TO 5] AND +path_parents:"\\/a\\/b")',
                        '+(+path_depth:[3 TO 4] AND +path_parents:"\\/c")',
                    ]
                )
            },
        )


class QueryParameterTests(TestCase):
    layer = COLLECTIVE_SOLR_MOCK_REGISTRY_FIXTURE

    def testSortIndex(self):
        extract = subtractQueryParameters
        params = extract({"sort_on": "foo"})
        self.assertEqual(params, dict(sort="foo asc"))
        # again with dashed instead of underscores
        params = extract({"sort-on": "foo"})
        self.assertEqual(params, dict(sort="foo asc"))

    def testSortOrder(self):
        extract = subtractQueryParameters
        params = extract({"sort_on": "foo", "sort_order": "ascending"})
        self.assertEqual(params, dict(sort="foo asc"))
        params = extract({"sort_on": "foo", "sort_order": "descending"})
        self.assertEqual(params, dict(sort="foo desc"))
        params = extract({"sort_on": "foo", "sort_order": "reverse"})
        self.assertEqual(params, dict(sort="foo desc"))
        params = extract({"sort_on": "foo", "sort_order": "bar"})
        self.assertEqual(params, dict(sort="foo asc"))
        # again with dashed instead of underscores
        params = extract({"sort-on": "foo", "sort-order": "ascending"})
        self.assertEqual(params, dict(sort="foo asc"))
        params = extract({"sort-on": "foo", "sort-order": "descending"})
        self.assertEqual(params, dict(sort="foo desc"))
        params = extract({"sort-on": "foo", "sort-order": "reverse"})
        self.assertEqual(params, dict(sort="foo desc"))
        params = extract({"sort-on": "foo", "sort-order": "bar"})
        self.assertEqual(params, dict(sort="foo asc"))

    def testSortLimit(self):
        extract = subtractQueryParameters
        params = extract({"sort_limit": 5})
        self.assertEqual(params, dict(rows=5))
        params = extract({"sort_limit": "10"})
        self.assertEqual(params, dict(rows=10))
        # again with dashed instead of underscores
        params = extract({"sort-limit": 5})
        self.assertEqual(params, dict(rows=5))
        params = extract({"sort-limit": "10"})
        self.assertEqual(params, dict(rows=10))

    def testBatchParameters(self):
        extract = subtractQueryParameters
        params = extract({"b_start": 5})
        self.assertEqual(params, dict(start=5))
        params = extract({"b_start": "10"})
        self.assertEqual(params, dict(start=10))
        params = extract({"b_size": 5})
        self.assertEqual(params, dict(rows=5))
        params = extract({"b_size": "10"})
        self.assertEqual(params, dict(rows=10))

    def testCombined(self):
        extract = subtractQueryParameters
        params = extract({"sort_on": "foo", "sort_limit": 5})
        self.assertEqual(params, dict(sort="foo asc", rows=5))
        params = extract({"sort_on": "foo", "sort_order": "reverse", "sort_limit": 5})
        self.assertEqual(params, dict(sort="foo desc", rows=5))
        params = extract({"sort_order": "reverse", "sort_limit": 5})
        self.assertEqual(params, dict(rows=5))

    def testAllowFacetParameters(self):
        extract = subtractQueryParameters
        # 'facet' and 'facet.*' should be passed on...
        params = extract({"facet": "true"})
        self.assertEqual(params, {"facet": "true"})
        params = extract({"facet.field": "foo", "facet.foo": "bar"})
        self.assertEqual(params, {"facet.field": "foo", "facet.foo": "bar"})
        params = extract({"facet.field": ("foo", "bar")})
        self.assertEqual(params, {"facet.field": ("foo", "bar")})
        # not 'facet*' though
        params = extract({"facetfoo": "bar"})
        self.assertEqual(params, {})
        # an underscore can be used instead of the '.' for conveniently
        # passing parameters as keyword arguments...
        params = extract(dict(facet_foo="bar"))
        self.assertEqual(params, {"facet.foo": "bar"})
        params = extract(dict(facet_foo=("foo", "bar")))
        self.assertEqual(params, {"facet.foo": ("foo", "bar")})

    def testAllowFilterQueryParameters(self):
        extract = subtractQueryParameters
        # 'fq' should be passed on...
        params = extract({"fq": "foo"})
        self.assertEqual(params, {"fq": "foo"})
        params = extract({"fq": ["foo", "bar"]})
        self.assertEqual(params, {"fq": ["foo", "bar"]})

    def testAllowFieldListParameter(self):
        extract = subtractQueryParameters
        # 'fl' should be passed on...
        params = extract({"fl": "foo"})
        self.assertEqual(params, {"fl": "foo"})
        params = extract({"fl": ["foo", "bar"]})
        self.assertEqual(params, {"fl": ["foo", "bar"]})

    def testAllowHighlightParameter(self):
        extract = subtractQueryParameters
        # 'hl' should be passed on...
        params = extract({"hl": "foo"})
        self.assertEqual(params, {"hl": "foo"})
        params = extract({"hl": ["foo", "bar"]})
        self.assertEqual(params, {"hl": ["foo", "bar"]})

    def testAllowRequestHandlerParamter(self):
        extract = subtractQueryParameters
        params = extract({"request_handler": "custom"})
        self.assertEqual(params, {"request_handler": "custom"})

    def testAllowDistanceParameter(self):
        # d=distance is a Spatial Filter
        extract = subtractQueryParameters
        # 'd' should be passed on...
        params = extract({"d": 0.2})
        self.assertEqual(params, {"d": 0.2})

    def testAllowPointParameter(self):
        # p=point is a Spatial Filter
        extract = subtractQueryParameters
        # 'p' should be passed on...
        params = extract({"pt": "50.73743,7.09821"})
        self.assertEqual(params, {"pt": "50.73743,7.09821"})

    def testAllowSfieldParameter(self):
        # sfield=spatial indexed field is a Spatial Filter
        extract = subtractQueryParameters
        # 'p' should be passed on...
        params = extract({"sfield": "geolocation"})
        self.assertEqual(params, {"sfield": "geolocation"})

    def testSortIndexCleanup(self):
        cleanup = cleanupQueryParameters
        schema = SolrSchema()
        # a non-existing sort index should be removed
        params = cleanup(dict(sort="foo asc"), schema)
        self.assertEqual(params, dict())
        # the same goes when the given index isn't indexed
        schema["foo"] = SolrField(indexed=False)
        params = cleanup(dict(sort="foo asc"), schema)
        self.assertEqual(params, dict())
        # a suitable index will be left intact, of course...
        schema["foo"].indexed = True
        params = cleanup(dict(sort="foo asc"), schema)
        self.assertEqual(params, dict(sort="foo asc"))
        # also make sure sort index aliases work, if the alias index exists
        params = cleanup(dict(sort="sortable_title asc"), schema)
        self.assertEqual(params, dict())
        schema["Title"] = SolrField(indexed=True)
        params = cleanup(dict(sort="sortable_title asc"), schema)
        self.assertEqual(params, dict(sort="Title asc"))

    def testFilterQuerySubstitution(self):
        def optimize(**params):
            query = dict(a="a:23", b="b:42", c="c:(23 42)")
            optimizeQueryParameters(query, params)
            return query, params

        # first test without the configuration utility
        self.assertEqual(optimize(), (dict(a="a:23", b="b:42", c="c:(23 42)"), dict()))
        # now unconfigured...
        config = getConfig()
        self.assertEqual(optimize(), (dict(a="a:23", b="b:42", c="c:(23 42)"), dict()))
        config.filter_queries = ["a"]
        self.assertEqual(optimize(), (dict(b="b:42", c="c:(23 42)"), dict(fq=["a:23"])))
        self.assertEqual(
            optimize(fq="x:13"),
            (dict(b="b:42", c="c:(23 42)"), dict(fq=["x:13", "a:23"])),
        )
        self.assertEqual(
            optimize(fq=["x:13", "y:17"]),
            (dict(b="b:42", c="c:(23 42)"), dict(fq=["x:13", "y:17", "a:23"])),
        )
        config.filter_queries = ["a", "c"]
        self.assertEqual(optimize(), (dict(b="b:42"), dict(fq=["a:23", "c:(23 42)"])))
        self.assertEqual(
            optimize(fq="x:13"),
            (dict(b="b:42"), dict(fq=["x:13", "a:23", "c:(23 42)"])),
        )
        self.assertEqual(
            optimize(fq=["x:13", "y:17"]),
            (dict(b="b:42"), dict(fq=["x:13", "y:17", "a:23", "c:(23 42)"])),
        )
        # also test substitution of combined filter queries
        config.filter_queries = ["a c"]
        self.assertTrue(
            optimize()
            in [
                (dict(b="b:42"), dict(fq=["c:(23 42) a:23"])),
                (dict(b="b:42"), dict(fq=["a:23 c:(23 42)"])),
            ]
        )
        config.filter_queries = ["a c", "b"]
        self.assertIn(
            optimize(),
            [
                ({"*": "*:*"}, dict(fq=["c:(23 42) a:23", "b:42"])),
                ({"*": "*:*"}, dict(fq=["a:23 c:(23 42)", "b:42"])),
            ],
        )
        # for multiple matches the first takes precedence
        config.filter_queries = ["a", "a c", "b"]
        self.assertEqual(optimize(), (dict(c="c:(23 42)"), dict(fq=["a:23", "b:42"])))
        # parameters not contained in the query must not be converted
        config.filter_queries = ["a nonexisting", "b"]
        self.assertEqual(optimize(), (dict(a="a:23", c="c:(23 42)"), dict(fq=["b:42"])))

    def testFilterFacetDependencies(self):
        extract = subtractQueryParameters
        # any info about facet dependencies must not be passed on to solr
        params = extract({"facet.field": "foo:bar", "facet.foo": "bar:foo"})
        self.assertEqual(params, {"facet.field": "foo", "facet.foo": "bar"})
        params = extract({"facet.field": ("foo:bar", "bar:foo")})
        self.assertEqual(params, {"facet.field": ("foo", "bar")})
        # also check the "underscore" variant...
        params = extract(dict(facet_foo="bar:foo"))
        self.assertEqual(params, {"facet.foo": "bar"})
        params = extract(dict(facet_foo=("foo:bar", "bar:foo")))
        self.assertEqual(params, {"facet.foo": ("foo", "bar")})


@mock.patch("collective.solr.stopword.raw", None)
@mock.patch("collective.solr.stopword.cooked", None)
class QueryManglerStopwordsTests(TestCase):
    layer = COLLECTIVE_SOLR_MOCK_REGISTRY_FIXTURE

    def setUp(self):
        self.config = getConfig()
        self.config.stopwords_case_insensitive = False
        self.config.stopwords = """
no
can
do
"""
        self.config.force_simple_search = True
        provideUtility(self.config, ISolrConnectionConfig)

    def tearDown(self):
        gsm = getGlobalSiteManager()
        gsm.unregisterUtility(self.config, ISolrConnectionConfig)

    def testBase(self):
        self.assertEqual(
            mangleSearchableText("beautiful world", self.config),
            "(beautiful* OR beautiful) (world* OR world)",
        )

    def testNoSimpleSearch(self):
        self.config.force_simple_search = False
        self.assertEqual(
            mangleSearchableText("beautiful can world", self.config),
            "(beautiful* OR beautiful) (can) (world* OR world)",
        )

    def testBaseSingleWord(self):
        self.assertEqual(
            mangleSearchableText("beautiful", self.config),
            "(beautiful* OR beautiful)",
        )

    def testRemovesWildcardTermSingleWord(self):
        self.assertEqual(
            mangleSearchableText("can", self.config),
            "(can)",
        )

    def testRemovesWildcardTermMultiWord(self):
        self.assertEqual(
            mangleSearchableText("beautiful can world", self.config),
            "(beautiful* OR beautiful) (can) (world* OR world)",
        )
        self.assertEqual(
            mangleSearchableText("no can do", self.config),
            "(no) (can) (do)",
        )

    def testRemovesWildcardCaseSensitiveUppercase(self):
        self.assertEqual(
            mangleSearchableText("beautiful Can world", self.config),
            "(beautiful* OR beautiful) (can* OR Can) (world* OR world)",
        )
        self.assertEqual(
            mangleSearchableText("No CAN dO", self.config),
            "(no* OR No) (can* OR CAN) (do* OR dO)",
        )

    def testRemovesWildcardCaseAgnosticUppercase(self):
        self.config.stopwords_case_insensitive = True
        self.assertEqual(
            mangleSearchableText("beautiful Can world", self.config),
            "(beautiful* OR beautiful) (Can) (world* OR world)",
        )
        self.assertEqual(
            mangleSearchableText("No CAN dO", self.config),
            "(No) (CAN) (dO)",
        )

    def testRemovesWildcardCaseSensitive(self):
        self.config.stopwords = """
No
CAN
dO
"""
        self.assertEqual(
            mangleSearchableText("beautiful Can world", self.config),
            "(beautiful* OR beautiful) (can* OR Can) (world* OR world)",
        )
        self.assertEqual(
            mangleSearchableText("nO can Do", self.config),
            "(no* OR nO) (can* OR can) (do* OR Do)",
        )

    def testRemovesWildcardCaseAgnostic(self):
        self.config.stopwords_case_insensitive = True
        self.config.stopwords = """
No
CAN
dO
"""
        self.assertEqual(
            mangleSearchableText("beautiful Can world", self.config),
            "(beautiful* OR beautiful) (Can) (world* OR world)",
        )
        self.assertEqual(
            mangleSearchableText("nO can Do", self.config),
            "(nO) (can) (Do)",
        )



from threading import local

# a thread-local object holding data for the queue
localData = local()
marker = []


def getLocal(name, factory=lambda: None):
    """get named thread-local value and optionally initialize it"""
    value = getattr(localData, name, marker)
    if value is marker:
        value = factory()
        setLocal(name, value)
    return value


def setLocal(name, value):
    """set a value for the named thread-local variable"""
    setattr(localData, name, value)



import json
import sys

from collective.solr.interfaces import IFlare, ISolrFlare
from collective.solr.parser import AttrDict
from DateTime import DateTime
from OFS.Traversable import path2url
from Products.CMFPlone.utils import pretty_title_or_id
from zope.component import adapts
from zope.component.hooks import getSite
from zope.interface import implementer
from zope.publisher.interfaces.http import IHTTPRequest

timezone = DateTime().timezone()


@implementer(IFlare)
class PloneFlare(AttrDict):
    """a sol(a)r brain, i.e. a data container for search results"""

    adapts(ISolrFlare, IHTTPRequest)

    __allow_access_to_unprotected_subobjects__ = True

    def __init__(self, context, request=None):
        self.context = context
        self.request = request
        self.update(context)  # copy data

    @property
    def id(self):
        """convenience alias"""
        return self.get("id", self.get("getId"))

    def getPath(self):
        """convenience alias"""
        return self["path_string"]

    def getRID(self):
        """Return a record id"""
        return int(int(self["UID"], 16) % sys.maxsize)

    def getObject(self, REQUEST=None, restricted=True):
        """return the actual object corresponding to this flare while
        mimicking what publisher's traversal does, i.e. potentially
        allowing access to the final object even if intermediate objects
        cannot be accessed (much like the original implementation in
        `ZCatalog.CatalogBrains.AbstractCatalogBrain`)"""
        site = getSite()
        path = self.getPath()
        if not path:
            return None
        path = path.split("/")
        if restricted:
            parent = site.unrestrictedTraverse(path[:-1])
            return parent.restrictedTraverse(path[-1])
        return site.unrestrictedTraverse(path)

    def _unrestrictedGetObject(self):
        return self.getObject(restricted=False)

    def getURL(self, relative=False):
        """convert the physical path into a url, if it was stored"""
        path = self.getPath()
        path = path
        try:
            url = self.request.physicalPathToURL(path, relative)
        except AttributeError:
            url = path2url(path.split("/"))
        return url

    def pretty_title_or_id(self):
        context = getSite()
        return pretty_title_or_id(context, self)

    @property
    def CreationDate(self):
        created = self.get("created", None)
        if created is None:
            return "n.a."
        return created.toZone(timezone).ISO8601()

    @property
    def ModificationDate(self):
        modified = self.get("modified", None)
        if modified is None:
            return "n.a."
        return modified.toZone(timezone).ISO8601()

    @property
    def data_record_normalized_score_(self):
        score = self.get("score", None)
        if score is None:
            return "n.a."
        return "%.1f" % (float(score) * 100)

    @property
    def review_state(self):
        if "review_state" in self:
            return self["review_state"]
        return ""

    @property
    def image_scales(self):
        if self.get("image_scales"):
            return json.loads(self["image_scales"])



from unittest import TestCase

from collective.solr.dispatcher import solrSearchResults
from collective.solr.testing import (
    LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING,
    activateAndReindex,
)
from plone.app.testing import SITE_OWNER_NAME, SITE_OWNER_PASSWORD
from plone.uuid.interfaces import IUUID

try:
    from plone.testing.zope import Browser
except ImportError:
    from plone.testing.z2 import Browser

from transaction import commit
from ZODB.POSException import ConflictError
from zope import component
from zope.interface import Interface
from zope.lifecycleevent.interfaces import IObjectCreatedEvent

UIDS = []


def raise_on_first_add(context, event):
    first = not bool(UIDS)
    UIDS.append(IUUID(context))
    if first:
        raise ConflictError()  # trigger a retry (once)


class ConflictTests(TestCase):

    layer = LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING

    def setUp(self):
        activateAndReindex(self.layer["portal"])
        commit()
        self.browser = Browser(self.layer["app"])
        self.browser.addHeader(
            "Authorization", "Basic %s:%s" % (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)
        )

    def test_retry_on_conflict(self):
        """This tests transaction handling when indexing in Solr, or more
        specifically properly aborting a transaction.  To do this we'll try to
        create some content and fake a `ConflictError` shortly before the
        transaction completes.  The publisher will catch it and retry, but
        while doing so the object will get a different UID than the first time.
        Without being able to abort the transaction Solr would receive two sets
        of data and consequently return two results when searching for this
        particular piece of content later on.
        """
        self.browser.open(self.layer["portal"].absolute_url())
        self.browser.getLink("Page").click()
        self.browser.getControl("Title", index=0).value = "Foo"
        component.provideHandler(raise_on_first_add, (Interface, IObjectCreatedEvent))
        self.browser.getControl("Save").click()
        self.assertEqual(len(UIDS), 2)
        self.assertEqual(len(solrSearchResults(SearchableText="Foo")), 1)

        sm = component.getSiteManager()
        sm.unregisterHandler(raise_on_first_add, (Interface, IObjectCreatedEvent))



from unittest import TestCase

from zope.component import getGlobalSiteManager, provideUtility

from collective.solr.interfaces import ISolrConnectionConfig
from collective.solr.testing import COLLECTIVE_SOLR_MOCK_REGISTRY_FIXTURE
from collective.solr.utils import getConfig
from collective.solr.stopword import isStopWord
import collective.solr.stopword
from unittest import mock


@mock.patch("collective.solr.stopword.raw", None)
@mock.patch("collective.solr.stopword.raw_case_insensitive", None)
@mock.patch("collective.solr.stopword.cooked", None)
class StopwordsUtilTests(TestCase):
    layer = COLLECTIVE_SOLR_MOCK_REGISTRY_FIXTURE

    def setUp(self):
        self.config = getConfig()
        self.config.force_simple_search = True
        self.config.stopwords_case_insensitive = False
        self.config.stopwords = """stopone
stoptwo
"""
        provideUtility(self.config, ISolrConnectionConfig)

    def tearDown(self):
        gsm = getGlobalSiteManager()
        gsm.unregisterUtility(self.config, ISolrConnectionConfig)

    def testIsStopword(self):
        self.assertTrue(isStopWord("stopone", self.config))
        self.assertTrue(isStopWord("stoptwo", self.config))
        self.assertFalse(isStopWord("stopthree", self.config))

    def testCachingPreserves(self):
        self.assertTrue(isStopWord("stopone", self.config))
        self.assertTrue(isStopWord("stoptwo", self.config))
        self.assertFalse(isStopWord("stopthree", self.config))
        collective.solr.stopword.cooked = ["stopone", "stopthree"]
        self.assertTrue(isStopWord("stopone", self.config))
        self.assertFalse(isStopWord("stoptwo", self.config))
        self.assertTrue(isStopWord("stopthree", self.config))

    def testCachingUpdates(self):
        self.assertTrue(isStopWord("stopone", self.config))
        self.assertTrue(isStopWord("stoptwo", self.config))
        self.assertFalse(isStopWord("stopthree", self.config))
        self.config.stopwords = """stopone
stopthree
"""
        self.assertTrue(isStopWord("stopone", self.config))
        self.assertFalse(isStopWord("stoptwo", self.config))
        self.assertTrue(isStopWord("stopthree", self.config))

    def testCachingCaseInsensitivityUpdates(self):
        self.config.stopwords = """Stopone
Stoptwo
"""
        self.config.stopwords_case_insensitive = True
        self.assertTrue(isStopWord("stopone", self.config))
        self.assertTrue(isStopWord("stoptwo", self.config))
        self.assertFalse(isStopWord("stopthree", self.config))
        self.config.stopwords_case_insensitive = False
        self.assertTrue(isStopWord("Stopone", self.config))
        self.assertTrue(isStopWord("Stoptwo", self.config))
        self.assertFalse(isStopWord("Stopthree", self.config))

    def testUnicode(self):
        self.config.stopwords = """für
daß
"""
        self.assertTrue(isStopWord("für", self.config))
        self.assertTrue(isStopWord("daß", self.config))

    def testCases(self):
        self.config.stopwords = """Uppercase
CamelCase
FULLCAPS
"""
        self.assertTrue(isStopWord("Uppercase", self.config))
        self.assertTrue(isStopWord("CamelCase", self.config))
        self.assertTrue(isStopWord("FULLCAPS", self.config))

    def testCaseInsensitive(self):
        self.config.stopwords = """Uppercase
CamelCase
FULLCAPS
"""
        self.config.stopwords_case_insensitive = True
        self.assertTrue(isStopWord("Uppercase", self.config))
        self.assertTrue(isStopWord("CamelCase", self.config))
        self.assertTrue(isStopWord("FULLCAPS", self.config))
        self.assertTrue(isStopWord("uppercase", self.config))
        self.assertTrue(isStopWord("camelcase", self.config))
        self.assertTrue(isStopWord("fullcaps", self.config))
        self.assertTrue(isStopWord("uppercasE", self.config))
        self.assertTrue(isStopWord("CAMELCASE", self.config))
        self.assertTrue(isStopWord("fullcaps", self.config))

    def testEmptyLines(self):
        self.config.stopwords = (
            """

stopone

"""
            + "   \n"
        )
        self.assertFalse(isStopWord("", self.config))
        self.assertFalse(isStopWord("   ", self.config))
        self.assertTrue(isStopWord("stopone", self.config))

    def testComments(self):
        self.config.stopwords = """stopone
stoptwo              | nostopone
| nostoptwo
stopthree            # nostopthree
# nostopfour
"""
        self.assertFalse(isStopWord("nostopone", self.config))
        self.assertFalse(isStopWord(" nostopone", self.config))
        self.assertFalse(isStopWord("nostoptwo", self.config))
        self.assertFalse(isStopWord(" nostoptwo", self.config))
        self.assertFalse(isStopWord("nostopthree", self.config))
        self.assertFalse(isStopWord(" nostopthree", self.config))
        self.assertFalse(isStopWord("nostopfour", self.config))
        self.assertFalse(isStopWord(" nostopfour", self.config))
        self.assertTrue(isStopWord("stopone", self.config))
        self.assertTrue(isStopWord("stoptwo", self.config))
        self.assertTrue(isStopWord("stopthree", self.config))

    def testLeadingSpaces(self):
        # stopwords.txt does not allow leading spaces, but the registry
        # pads it up because of the way we define it in the xml.
        self.config.stopwords = (
            """
  stopone
    stoptwo
"""
            + "   \n"
        )
        self.assertFalse(isStopWord("", self.config))
        self.assertFalse(isStopWord("  ", self.config))
        self.assertFalse(isStopWord("    ", self.config))
        self.assertTrue(isStopWord("stopone", self.config))
        self.assertTrue(isStopWord("stoptwo", self.config))



from datetime import date, datetime
from logging import getLogger
from socket import error

import json
import six
from Acquisition import aq_get
from collective.solr.exceptions import SolrConnectionException
from collective.solr.interfaces import (
    ICheckIndexable,
    ISolrAddHandler,
    ISolrConnectionManager,
    ISolrIndexQueueProcessor,
)
from collective.solr.utils import getConfig, prepareData
from DateTime import DateTime
from lxml import etree
from plone.indexer.interfaces import IIndexableObject, IIndexableObjectWrapper
from plone.registry.interfaces import IRegistry
from Products.CMFCore.CMFCatalogAware import CMFCatalogAware
from Products.CMFCore.utils import getToolByName
from requests_toolbelt import MultipartEncoder
from six.moves.urllib.parse import urlencode
from ZODB.interfaces import BlobError
from ZODB.POSException import ConflictError
from zope.component import (
    adapts,
    getUtility,
    queryAdapter,
    queryMultiAdapter,
    queryUtility,
)
from zope.interface import Interface, implementer

logger = getLogger("collective.solr.indexer")


@implementer(ICheckIndexable)
class BaseIndexable(object):

    adapts(Interface)

    def __init__(self, context):
        self.context = context

    def __call__(self):
        return isinstance(self.context, CMFCatalogAware)


def datehandler(value):
    # TODO: we might want to handle datetime and time as well;
    # check the enfold.solr implementation
    if value is None or value == "":
        raise AttributeError
    if isinstance(value, str) and not value.endswith("Z"):
        try:
            value = DateTime(value)
        except SyntaxError:
            raise AttributeError

    if isinstance(value, DateTime):
        v = value.toZone("UTC")
        value = "%04d-%02d-%02dT%02d:%02d:%02d.%03dZ" % (
            v.year(),
            v.month(),
            v.day(),
            v.hour(),
            v.minute(),
            int(v.second()),
            v.millis() % 1000,
        )
    elif isinstance(value, datetime):
        # Convert a timezone aware timetuple to a non timezone aware time
        # tuple representing utc time. Does nothing if object is not
        # timezone aware
        millis = int(
            value.microsecond / 1000
        )  # 1 second = 1 000 millis = 1 000 000 microsecond
        value = datetime(*value.utctimetuple()[:6])
        value = "%s.%03dZ" % (value.strftime("%Y-%m-%dT%H:%M:%S"), millis)
    elif isinstance(value, date):
        value = "%s.000Z" % value.strftime("%Y-%m-%dT%H:%M:%S")
    return value


def inthandler(value):
    if value is None or value == "":
        raise AttributeError("Solr cant handle none strings or empty values")
    else:
        return value


handlers = {
    "solr.DateField": datehandler,
    "solr.DatePointField": datehandler,
    "solr.FloatField": inthandler,
    "solr.FloatPointField": inthandler,
    "solr.TrieDateField": datehandler,
    "solr.TrieIntField": inthandler,
    "solr.IntField": inthandler,
    "solr.IntPointField": inthandler,
}


@implementer(ISolrAddHandler)
class DefaultAdder(object):
    """"""

    def __init__(self, context):
        self.context = context

    def __call__(self, conn, **data):
        # remove in Plone unused field links,
        # which gives problems with some documents
        data.pop("links", "")
        conn.add(**data)


class BinaryAdder(DefaultAdder):
    """Add binary content to index via tika"""

    def getblob(self):
        field = self.context.getPrimaryField()
        # do not return anything if a content object does not provide a primary field
        if not field:
            return
        return field.get(self.context).blob

    def getpath(self):
        blob = self.getblob()
        if blob is None:
            return None
        try:
            path = blob.committed()
        except BlobError:
            path = blob._p_blob_committed or blob._p_blob_uncommitted
        logger.debug("Indexing BLOB from path %s", path)
        return path

    def __call__(self, conn, **data):
        postdata = {}
        path = self.getpath()
        if path is None:
            super(BinaryAdder, self).__call__(conn, **data)

        postdata["extractFormat"] = "text"
        postdata["extractOnly"] = "true"
        postdata["wt"] = "xml"
        headers = conn.formheaders

        registry = getUtility(IRegistry)
        use_tika = registry.get("collective.solr.use_tika")
        tika_default_field = registry.get("collective.solr.tika_default_field")
        if not use_tika:
            # fall back to use SearchableText if tika is not enabled
            tika_default_field = "SearchableText"

        # blobs are accessed via the file system
        if use_tika:
            blob = self.getblob()
            if blob is None:
                return
            openedBlob = blob.open()

            postdata["myfile"] = (
                data["id"],
                openedBlob,
                data.get("content_type", "application/octet-stream"),
            )

            encodedPost = MultipartEncoder(fields=postdata)

            headers["Content-Type"] = encodedPost.content_type
            postdata_urlencoded = encodedPost.to_string()
        else:
            postdata["stream.file"] = self.getpath()
            postdata["stream.contentType"] = data.get(
                "content_type", "application/octet-stream"
            )
            postdata_urlencoded = urlencode(postdata, doseq=True)

        url = "%s/update/extract" % conn.solrBase

        try:
            response = conn.doPost(url, postdata_urlencoded, headers)
            root = etree.parse(response)
            data[tika_default_field] = root.find(".//str").text.strip()
        except SolrConnectionException as e:
            logger.warn("Error %s @ %s", e, data["path_string"])
            data[tika_default_field] = ""
        except etree.XMLSyntaxError as e:
            logger.warn("Parsing error %s @ %s.", e, data["path_string"])
            data[tika_default_field] = ""
        finally:
            if use_tika:
                openedBlob.close()

        super(BinaryAdder, self).__call__(conn, **data)


class DXFileBinaryAdder(BinaryAdder):

    fieldname = "file"

    def getblob(self):
        field = getattr(self.context, self.fieldname, None)
        return getattr(field, "_blob", None)


class DXImageBinaryAdder(DXFileBinaryAdder):

    fieldname = "image"


def boost_values(obj, data):
    """calculate boost values using a method or skin script;  returns
    a dictionary with the values or `None`"""
    boost_index_getter = aq_get(obj, "solr_boost_index_values", None)
    if boost_index_getter is not None:
        return boost_index_getter(data)


@implementer(ISolrIndexQueueProcessor)
class SolrIndexProcessor(object):
    """a queue processor for solr"""

    def __init__(self, manager=None):
        self.manager = manager

    def index(self, obj, attributes=None):
        """Index the specified attributes for obj using atomic updates, or all
        of them if `attributes` is `None`.
        Also make sure the `uniqueKey` is part of attributes, and passing the
        attributes to the self.getData() call to avoid causing Plone to index
        all fields instead of just the necessary ones.
        """
        conn = self.getConnection()
        if conn is not None and ICheckIndexable(obj)():
            schema = self.manager.getSchema()
            if schema is None:
                msg = "unable to fetch schema, skipping indexing of %r"
                logger.warning(msg, obj)
                return
            uniqueKey = schema.get("uniqueKey", None)
            if uniqueKey is None:
                msg = "schema is missing unique key, skipping indexing of %r"
                logger.warning(msg, obj)
                return

            if attributes:

                if "path" in attributes:
                    attributes = list(attributes)
                    attributes.extend(["path_string", "path_parents", "path_depth"])

                attributes = set(schema.keys()).intersection(attributes)
                if not attributes:
                    return

                if uniqueKey not in attributes:
                    # The uniqueKey is required in order to identify the
                    # document when doing atomic updates.
                    attributes.add(uniqueKey)

            data, missing = self.getData(obj, attributes=attributes)
            if not data:
                return  # don't index with no data...
            prepareData(data)
            if data.get(uniqueKey, None) is not None and not missing:
                registry = getUtility(IRegistry)
                config_commit_within = registry["collective.solr.commit_within"]  # noqa
                if config_commit_within:
                    data["commitWithin"] = config_commit_within
                try:
                    logger.debug("indexing %r (%r)", obj, data)
                    pt = data.get("portal_type", "default")
                    logger.debug("indexing %r with %r adder (%r)", obj, pt, data)

                    adder = queryAdapter(obj, ISolrAddHandler, name=pt)

                    if adder is None:
                        adder = DefaultAdder(obj)
                    adder(conn, boost_values=boost_values(obj, data), **data)
                except (SolrConnectionException, error):
                    logger.exception("exception during indexing %r", obj)

    def reindex(self, obj, attributes=None, update_metadata=False):
        if not attributes:
            attributes = None
        self.index(obj, attributes)

    def unindex(self, obj):
        conn = self.getConnection()
        if conn is not None:
            schema = self.manager.getSchema()
            if schema is None:
                msg = "unable to fetch schema, skipping unindexing of %r"
                logger.warning(msg, obj)
                return
            uniqueKey = schema.get("uniqueKey", None)
            if uniqueKey is None:
                msg = "schema is missing unique key, skipping unindexing of %r"
                logger.warning(msg, obj)
                return

            # remove the PathWrapper, otherwise IndexableObjectWrapper fails
            # to get the UID indexer (for dexterity objects) and the parent
            # UID is acquired
            if hasattr(obj, "context"):
                obj = obj.context

            data, missing = self.getData(obj, attributes=[uniqueKey])
            prepareData(data)
            if uniqueKey not in data:
                msg = "Can not unindex: no unique key for object %r"
                logger.info(msg, obj)
                return
            data_key = data[uniqueKey]
            if data_key is None:
                msg = "Can not unindex: `None` unique key for object %r"
                logger.info(msg, obj)
                return
            try:
                logger.debug("unindexing %r (%r)", obj, data)
                conn.delete(id=data_key)
            except (SolrConnectionException, error):
                logger.exception("exception during unindexing %r", obj)

    def begin(self):
        pass

    def commit(self, wait=None):
        conn = self.getConnection()
        if conn is not None:
            config = getConfig()
            if not isinstance(wait, bool):
                wait = not config.async_indexing
            try:
                logger.debug("committing")
                if not config.auto_commit or config.commit_within:
                    # If we have commitWithin enabled, we never want to do
                    # explicit commits. Even though only add's support this
                    # and we might wait a bit longer on delete's this way
                    conn.flush()
                else:
                    conn.commit(waitSearcher=wait)
            except (SolrConnectionException, error):
                logger.exception("exception during commit")
            self.manager.closeConnection()

    def abort(self):
        conn = self.getConnection()
        if conn is not None:
            logger.debug("aborting")
            conn.abort()
            self.manager.closeConnection()

    # helper methods

    def getConnection(self):
        if self.manager is None:
            self.manager = queryUtility(ISolrConnectionManager)
        if self.manager is not None:
            self.manager.setIndexTimeout()
            return self.manager.getConnection()

    def wrapObject(self, obj):
        """wrap object with an "IndexableObjectWrapper` (for Plone < 3.3) or
        adapt it to `IIndexableObject` (for Plone >= 3.3), see
        `CMFPlone...CatalogTool.catalog_object` for some background"""
        wrapper = obj
        # first try the new way, i.e. using `plone.indexer`...
        catalog = getToolByName(obj, "portal_catalog", None)
        adapter = queryMultiAdapter((obj, catalog), IIndexableObject)
        if adapter is not None:
            wrapper = adapter
        else:  # otherwise try the old way...
            portal = getToolByName(obj, "portal_url", None)
            if portal is None:
                return obj
            portal = portal.getPortalObject()
            adapter = queryMultiAdapter((obj, portal), IIndexableObjectWrapper)
            if adapter is not None:
                wrapper = adapter
            wft = getToolByName(obj, "portal_workflow", None)
            if wft is not None:
                wrapper.update(wft.getCatalogVariablesFor(obj))
        return wrapper

    def getData(self, obj, attributes=None):
        schema = self.manager.getSchema()
        if schema is None:
            return {}, ()
        if attributes is None:
            attributes = list(schema.keys())
        obj = self.wrapObject(obj)
        data = {}
        for name in attributes:
            try:
                value = getattr(obj, name)
                if callable(value):
                    value = value()
            except ConflictError:
                raise
            except AttributeError:
                continue
            except Exception:
                logger.exception("Error occured while getting data for indexing!")
                continue
            if isinstance(value, set):
                value = list(value)
            field = schema[name]
            handler = handlers.get(field.class_, None)
            if handler is not None:
                try:
                    value = handler(value)
                except AttributeError:
                    continue
            elif isinstance(value, (list, tuple)) and not field.multiValued:
                separator = getattr(field, "separator", " ")
                value = separator.join(value)
            if isinstance(value, six.binary_type):
                value = value.decode("utf-8")
            if name == "image_scales":
                # image_scales is stored in Plone as a PersistentMapping
                value = json.dumps(dict(value))
            data[name] = value
        missing = set(schema.requiredFields) - set(data.keys())
        return data, missing



# http://svn.apache.org/repos/asf/lucene/solr/trunk/client/python/solr.py

# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# A simple Solr client for python.
#
# quick examples on use:
#
# from collective.solr.solr import *
# c = SolrConnection(host='localhost:8983', persistent=True)
# c.add(id='500',name='python test doc')
# c.commit()
# c.search(q='id:[* TO *]', wt='python', rows='10',indent='on')
# c.delete('123')
# c.commit()

import base64
import codecs
import socket
from copy import deepcopy
from logging import getLogger
from xml.etree.cElementTree import fromstring
from xml.sax.saxutils import escape

import six
import six.moves.http_client
import six.moves.urllib.parse
import six.moves.urllib.request
from plone.dexterity.utils import safe_unicode
from requests_toolbelt import MultipartEncoder

from collective.solr.exceptions import (
    RepeatableSolrConnectionException,
    SolrConnectionException,
)
from collective.solr.parser import SolrSchema
from collective.solr.utils import getConfig, translation_map

logger = getLogger(__name__)


class SolrConnection:
    def __init__(
        self,
        host="localhost:8983",
        solrBase="/solr/plone",
        persistent=True,
        postHeaders={},
        timeout=None,
        login=None,
        password=None,
        manager=None,
    ):
        self.manager = manager
        self.host = host
        self.solrBase = str(solrBase)
        self.persistent = persistent
        self.reconnects = 0
        self.encoder = codecs.getencoder("utf-8")
        # responses from Solr will always be in UTF-8
        self.decoder = codecs.getdecoder("utf-8")
        # a real connection to the server is not opened at this point.
        self.conn = six.moves.http_client.HTTPConnection(self.host, timeout=timeout)
        # self.conn.set_debuglevel(1000000)
        self.login = login
        self.auth_headers = {}
        if login and password:
            credentials = "{0}:{1}".format(login, password)
            authorization = base64.b64encode(credentials.encode("ascii"))
            self.auth_headers["Authorization"] = "Basic {0}".format(
                authorization.decode("ascii")
            )
        self.xmlbody = []
        self.xmlheaders = {"Content-Type": "text/xml; charset=utf-8"}
        self.xmlheaders.update(postHeaders)
        self.xmlheaders.update(self.auth_headers)
        if not self.persistent:
            self.xmlheaders["Connection"] = "close"
        self.formheaders = {
            "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
        }
        self.formheaders.update(self.auth_headers)
        if not self.persistent:
            self.formheaders["Connection"] = "close"

    def __str__(self):
        xmlheaders = deepcopy(self.xmlheaders)
        if "Authorization" in xmlheaders:
            xmlheaders["Authorization"] = "Basic ***"
        return (
            "SolrConnection{host=%s, solrBase=%s, persistent=%s, "
            "postHeaders=%s, reconnects=%s, login=%s, password=%s}"
            % (
                self.host,
                self.solrBase,
                self.persistent,
                xmlheaders,
                self.reconnects,
                self.login,
                self.auth_headers and "***" or None,
            )
        )

    def __reconnect(self):
        self.reconnects += 1
        self.conn.close()
        self.conn.connect()

    reset = __reconnect

    def close(self):
        self.conn.close()

    def __errcheck(self, rsp):
        if rsp.status != 200:
            body = None
            try:
                body = rsp.read()
            except:  # noqa
                pass
            if (
                rsp.status == 500
                and body is not None
                and b"Missing initial multi part boundary" in body
            ):
                raise RepeatableSolrConnectionException(rsp.status, rsp.reason, body)
            else:
                raise SolrConnectionException(rsp.status, rsp.reason, body)
        return rsp

    def setTimeout(self, timeout):
        """set a timeout value for the currently open connection"""
        logger.debug("setting socket timeout on %r: %s", self, timeout)
        self.conn.timeout = timeout

    def doPost(self, url, body, headers):
        return self.doGetOrPost("POST", url, body, headers)

    def doGet(self, url, headers):
        return self.doGetOrPost("GET", url, "", headers)

    def doGetOrPost(self, method, url, body, headers):
        if not isinstance(body, (six.binary_type, MultipartEncoder)):
            body = body.encode("utf-8")
        try:
            self.conn.request(method, url, body, headers)
            return self.__errcheck(self.conn.getresponse())
        except (
            socket.error,
            six.moves.http_client.CannotSendRequest,
            six.moves.http_client.ResponseNotReady,
            six.moves.http_client.BadStatusLine,
        ):
            # Reconnect in case the connection was broken from the server
            # going down, the server timing out our persistent connection, or
            # another network failure. Also catch httplib.CannotSendRequest,
            # httlib.ResponseNotReady and httlib.BadStatusLine because the
            # HTTPConnection object can get in a bad state (seems like they
            # might be "ghosted" in the zodb).
            self.__reconnect()
            self.conn.request(method, url, body, headers)
            return self.__errcheck(self.conn.getresponse())

    def doUpdateXML(self, request):
        # solr will support abort/rollback only from version 1.4, so
        # for now we delay sending the xml until the commit...
        # see http://issues.apache.org/jira/browse/SOLR-670
        logger.debug("storing xml request for later: %r", request)
        self.xmlbody.append(request)

    def flush(self):
        """send out the stored requests to solr"""
        count = 0
        responses = []
        for request in self.xmlbody:
            try:
                responses.append(self.doSendXML(request))
            except (SolrConnectionException, socket.error):
                logger.exception("exception during request %r", request)
            count += len(request)
        logger.debug("flushed out %d bytes in %d requests", count, len(self.xmlbody))
        del self.xmlbody[:]
        return responses

    def doSendXML(self, request):
        try:
            rsp = self.doPost(self.solrBase + "/update", request, self.xmlheaders)
            data = rsp.read()
        finally:
            if not self.persistent:
                self.conn.close()
        # detect old-style error response (HTTP response code of
        # 200 with a non-zero status.
        parsed = fromstring(self.decoder(data)[0])
        status = parsed.attrib.get("status", 0)
        if status != 0:
            reason = parsed.documentElement.firstChild.nodeValue
            raise SolrConnectionException(rsp.status, reason)
        return parsed

    def escapeVal(self, val):
        if not isinstance(val, six.text_type):
            try:
                val = six.text_type(val)
            except UnicodeDecodeError:
                val = safe_unicode(val)
        if six.PY2:
            val = val.encode("utf-8")
        escaped_val = escape(val.translate(translation_map))
        if isinstance(escaped_val, six.binary_type):
            escaped_val = escaped_val.decode("utf-8")
        return escaped_val

    def escapeKey(self, key):
        if not isinstance(key, six.text_type):
            key = six.text_type(key)
        key = key.replace("&", "&amp;")
        key = key.replace('"', "&quot;")
        return key

    def delete(self, id):
        xstr = "<delete><id>%s</id></delete>" % self.escapeVal(id)
        return self.doUpdateXML(xstr)

    def deleteByQuery(self, query):
        xstr = "<delete><query>%s</query></delete>" % self.escapeVal(query)
        return self.doUpdateXML(xstr)

    def get_schema(self):
        """Memoized access to Solr Schema.

        We need to know which one is the uniqueKey in the add() method below,
        but don't want to hit Solr with a request for the schema on every
        single add().
        """
        if not hasattr(self, "_schema"):
            self._schema = self.getSchema()
        return self._schema

    def add(self, boost_values=None, atomic_updates=True, **fields):
        solr_config = getConfig()
        atomic_updates_enabled = getattr(solr_config, "atomic_updates", atomic_updates)

        schema = self.get_schema()
        uniqueKey = schema.get("uniqueKey", None)
        if uniqueKey is None:
            raise Exception("Could not get uniqueKey from Solr schema")

        if uniqueKey not in fields:
            logger.warning(
                "uniqueKey '%s' missing for item %s, skipping" % (uniqueKey, fields)
            )
            return

        latlon_fields = [
            field["name"]
            for field in schema.fields
            if field["class"] == "solr.LatLonPointSpatialField"
        ]

        within = fields.pop("commitWithin", None)
        if within:
            lst = ['<add commitWithin="%s">' % str(within)]
        else:
            lst = ["<add>"]
        if boost_values is None:
            boost_values = {}
        if "" in boost_values:  # boost value for the entire document
            lst.append('<doc boost="%s">' % boost_values[""])
        else:
            lst.append("<doc>")

        for f, v in fields.items():
            # Add update="set" attribute to each field except for the uniqueKey
            # field.
            if f == uniqueKey:
                tmpl = '<field name="%s">%%s</field>' % self.escapeKey(f)
                lst.append(tmpl % self.escapeVal(v))
                continue

            # geolocation need to be explicitly deleted. They can't index
            # None or empty strings.
            if f in latlon_fields and not v:
                tmpl = '<field name="%s" update="set" null="true"/>'
                lst.append(tmpl % (self.escapeKey(f)))
                continue
            if f in boost_values:
                tmpl = '<field name="%s" boost="%s" update="set">%%s</field>'
                tmpl = tmpl % (self.escapeKey(f), boost_values[f])
            else:
                tmpl = '<field name="%s" update="set">%%s</field>'
                tmpl = tmpl % self.escapeKey(f)

            if not atomic_updates_enabled:
                # Remove update="set", since it breaks the index time boosting.
                tmpl = tmpl.replace(' update="set"', "")

            if isinstance(v, (list, tuple)):  # multi-valued
                for value in v:
                    lst.append(tmpl % self.escapeVal(value))
                if not v:
                    tmpl = '<field name="%s" update="set" null="true"/>'
                    lst.append(tmpl % (self.escapeKey(f)))
            else:
                if v is None:
                    if f in boost_values:
                        tmpl = '<field name="%s" boost="%s" update="set" null="true"/>'
                        tmpl = tmpl % (self.escapeKey(f), boost_values[f])
                    else:
                        tmpl = '<field name="%s" update="set" null="true"/>'
                        tmpl = tmpl % (self.escapeKey(f))
                    lst.append(tmpl)
                else:
                    lst.append(tmpl % self.escapeVal(v))
        lst.append("</doc>")
        lst.append("</add>")
        xstr = "".join(lst)

        if self.conn.debuglevel > 0:
            logger.info("Update message:\n" + xstr)

        return self.doUpdateXML(xstr)

    def commit(self, waitSearcher=True, optimize=False, soft=False):
        data = {
            "committype": optimize and "optimize" or "commit",
            "nowait": not waitSearcher and ' waitSearcher="false"' or "",
            "soft": soft and ' softCommit="true"' or "",
        }
        xstr = "<%(committype)s%(soft)s%(nowait)s/>" % data
        self.doUpdateXML(xstr)
        return self.flush()

    def abort(self):
        # solr will support abort/rollback only from version 1.4, so
        # for now we delay sending the xml until the commit (see above),
        # which is why we don't have to send anything to abort...
        # see http://issues.apache.org/jira/browse/SOLR-670
        logger.debug("aborting %d requests: %r", len(self.xmlbody), self.xmlbody)
        del self.xmlbody[:]

    def _search(self, **params):
        request_handler = params.get("request_handler", "select")
        if "request_handler" in params:
            del params["request_handler"]
        for key, value in params.items():
            if isinstance(value, six.text_type):
                params[key] = value.encode("utf-8")
        request = six.moves.urllib.parse.urlencode(params, doseq=True)
        logger.debug("sending request: %s" % request)
        try:
            response = self.doPost(
                "%s/%s" % (self.solrBase, request_handler), request, self.formheaders
            )
        finally:
            if not self.persistent:
                self.conn.close()
        return response

    def search(self, **params):
        try:
            return self._search(**params)
        except RepeatableSolrConnectionException:
            # Handle "Missing initial multi part boundary" errors, since
            # they are also caused by the connection stuck in the thread.
            # Not fixable by a reconnect, a new connection object has to be made.
            self.manager.closeConnection()
            duplicate = self.manager.getConnection()
            duplicate.reconnects = self.reconnects + 1
            return duplicate._search(**params)

    def getSchema(self):
        schema_url = "%s/admin/file/?file=schema.xml"
        # schema_url = '%s/admin/file/?file=managed-schema'
        logger.debug("getting schema from: %s", schema_url % self.solrBase)
        try:
            self.conn.request(
                "GET", schema_url % self.solrBase, headers=self.auth_headers
            )
            response = self.conn.getresponse()
        except (
            socket.error,
            six.moves.http_client.CannotSendRequest,
            six.moves.http_client.ResponseNotReady,
            six.moves.http_client.BadStatusLine,
        ):
            # see `doPost` method for more info about these exceptions
            self.__reconnect()
            self.conn.request(
                "GET", schema_url % self.solrBase, headers=self.auth_headers
            )
            response = self.conn.getresponse()

        if response.status == 200:
            xml = response.read().decode("utf-8")
            return SolrSchema(xml.strip())

        self.__errcheck(response)  # raise a SolrConnectionException



from unittest import TestCase

import transaction
from collective.solr.testing import LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING
from collective.solr.utils import activate
from plone import api
from plone.app.testing import TEST_USER_ID, setRoles
from plone.dexterity.utils import safe_unicode


class TestDublincoreSubjectField(TestCase):

    layer = LEGACY_COLLECTIVE_SOLR_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        subjects = [safe_unicode("\xc3\xbcber Keyword"), safe_unicode("Second one")]

        self.content = api.content.create(
            container=api.portal.get(),
            type="Document",
            title="test document",
            subject=subjects,
        )

    def test_successfully_reindex_umlauts(self):
        activate()
        self.content.reindexObject()
        # This failed before with a UnicodeDecodeError
        transaction.commit()
